"""
This module contains the Visitor class.

The Visitor class contains a decorator ``visitor`` that is intended to be
used as a decorator for methods.

Example:
    >>> class TypeCollector:
    ...     visitor = Visitor().visitor
    ...
    ...     @visitor
    ...     def visit(self, node: Program):
    ...         # some code
    ...
    ...     @visitor
    ...     def visit(self, node: Function):
    ...         # some code
"""

from typing import Callable, Dict, Tuple, get_type_hints


class Visitor:
    """
    Visitor class.

    A visitor instance must be creater for each class that will decorate
    methods with the ``visitor`` decorator.
    """

    def __init__(self):
        self.cases: Dict[Tuple[type], Callable] = {}
        self.callbacks = []

    def visitor(self, func):
        """
        Decorator for methods.

        The decorator is intended to be used as a decorator for methods.
        The method will be registered in the ``cases`` dictionary.
        """
        case_key = tuple(get_type_hints(func).values())
        if case_key in self.cases:
            raise ValueError(f"Duplicate visitor: {case_key}")

        def wrapper(*args, **kwargs):
            case_key = tuple(type(arg) for arg in args[1:])
            if case_key not in self.cases:
                raise ValueError(f"No {type(args[0]).__name__} visitor for {case_key}")
            case_func = self.cases[case_key]
            for callback in self.callbacks:
                callback(*args, **kwargs)
            return case_func(*args, **kwargs)

        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        self.cases[case_key] = func
        return wrapper

    def callback(self, callback):
        """
        Add a callback to the visitor.

        The callback will be called for each node that is visited.
        """
        self.callbacks.append(callback)
        return callback



from __future__ import annotations

from typing import Any, Callable, Dict, List


class Type:
    cs_types = {}

    def __init__(self, type_name: str, parent: Type = None):
        self.type_name = type_name
        self.attributes: Dict[str, Any] = {}
        self.parent = parent
        Type.cs_types[type_name] = self

    def add_attribute(self, attribute: str, default: Any = None):
        self.attributes[attribute] = default

    def method(self, method_name: str):
        def method_wrapper(func):
            self.add_attribute(method_name, func)
            return func

        return method_wrapper

    def get_attribute(self, attribute_name: str):
        for attribute in self.attributes:
            if attribute.name == attribute_name:
                return attribute
        return None

    def get_attr_dict(self):
        all_attrs = {}
        if self.parent:
            all_attrs.update(self.parent.get_attr_dict())
        all_attrs.update(self.attributes)
        return all_attrs

    def subtype(self, other: Type):
        if self.type_name == "any":
            return True
        if not isinstance(other, tuple):
            other = (other,)
        for subtype in other:
            if self.type_name == subtype.type_name:
                return True
        if self.parent is None:
            return False
        return self.parent.subtype(other)

    def __call__(self, value):
        return Instance(self, value)

    @staticmethod
    def new(type_name: str, *args, **kwargs):
        if type_name not in Type.cs_types:
            raise ValueError(f"{type_name} is not a valid Clismo type")
        return Type.cs_types[type_name](*args, **kwargs)

    @staticmethod
    def get(type_name: str):
        if type_name == "list":
            return (
                cs_list_of_float,
                cs_list_of_int,
                cs_list_of_str,
                cs_list_of_bool,
                cs_list_of_client,
                cs_list_of_server,
                cs_list_of_step,
                cs_list_of_simulation,
            )
        if type_name not in Type.cs_types:
            raise ValueError(f"{type_name} is not a valid Clismo type")
        return Type.cs_types[type_name]

    @staticmethod
    def get_type(value):
        if isinstance(value, str):
            return Type.get("str")
        if isinstance(value, bool):
            return Type.get("bool")
        if isinstance(value, int):
            return Type.get("int")
        if isinstance(value, float):
            return Type.get("float")
        if value is None:
            return Type.get("none")
        if isinstance(value, dict):
            return Type.get("dict")
        if isinstance(value, tuple):
            return Type.get("tuple")
        if callable(value):
            return Type.get("function")
        return value

    @staticmethod
    def resolve_type(value):
        val_type = Type.get_type(value)
        return val_type(value)

    def __repr__(self):
        return f"<Type {self.type_name}>"

    def __str__(self):
        return f"<Type {self.type_name}>"


class Instance:
    def __init__(self, _type: Type, value):
        self.type = _type
        self.value = value


cs_object = Type("object")
cs_float = Type("float", cs_object)
cs_int = Type("int", cs_float)
cs_bool = Type("bool", cs_int)
cs_str = Type("str", cs_object)
cs_none = Type("none", cs_object)
cs_any = Type("any", cs_object)
cs_client = Type("client", cs_object)
cs_server = Type("server", cs_object)
cs_step = Type("step", cs_object)
cs_simulation = Type("simulation", cs_object)

cs_list_of_float = Type("float_list", cs_object)
cs_list_of_int = Type("int_list", cs_object)
cs_list_of_str = Type("str_list", cs_object)
cs_list_of_bool = Type("bool_list", cs_object)
cs_list_of_client = Type("client_list", cs_object)
cs_list_of_server = Type("server_list", cs_object)
cs_list_of_step = Type("step_list", cs_object)
cs_list_of_simulation = Type("simulation_list", cs_object)



# import random
# from contextlib import redirect_stdout
# from time import time

# #import clismo.nl_ast as ast
# from clismo.ia.genetic_alg import GeneticAlg
# from clismo.lang.context import Context
# #from clismo.visitors.eval_visitor import EvalVisitor
# #from clismo.visitors.opt_visitor import OptVisitor


# class CodeOptimizer(GeneticAlg):
#     """
#     Genetic algorithm for optimizing the code
#     """

#     def __init__(
#         self,
#         ast: ast.AST,
#         possible_changes: list,
#         max_iter: int = 5,
#         minimize=True,
#         population_size=10,
#         mutation_prob=0.1,
#         best_selection_count=3,
#         generate_new_randoms=2,
#     ):
#         self.ast = ast
#         opt = OptVisitor()
#         opt.check(self.ast)
#         self.possible_changes = opt.changes
#         self.last_vector = [0] * len(possible_changes)
#         super().__init__(
#             max_iter,
#             minimize,
#             population_size,
#             mutation_prob,
#             best_selection_count,
#             generate_new_randoms,
#         )

#     def eval(self, solution) -> float:
#         """
#         Evaluate the fitness of the solution
#         """
#         for i, item in enumerate(solution):
#             if self.last_vector[i] != item:
#                 node, change_func, revert_func = self.possible_changes[i]
#                 func = change_func if item else revert_func
#                 func(node)

#         evaluator = EvalVisitor(Context())
#         start = time()
#         try:
#             with open("logs.txt", "w+") as f:
#                 with redirect_stdout(f):
#                     evaluator.eval(self.ast)
#         except Exception as e:
#             if isinstance(e, KeyboardInterrupt):
#                 raise e
#             self.last_vector = solution
#             return float("inf") if self.minimize else 0
#         end = time()
#         self.last_vector = solution
#         return end - start

#     def get_random_solution(self):
#         """
#         Generate a random solution
#         """
#         vals = [random.randint(0, 1) for _ in range(len(self.possible_changes))]
#         return vals

#     def crossover(self, sol1, sol2):
#         """
#         Crossover two solutions
#         """
#         idx = random.randint(0, len(self.possible_changes) - 1)
#         return sol1[:idx] + sol2[idx:], sol2[:idx] + sol1[idx:]

#     def mutate(self, sol):
#         """
#         Mutate a solution
#         """
#         idx = random.randint(0, len(self.possible_changes) - 1)
#         sol[idx] = 1 - sol[idx]
#         return sol



import logging
from typing import Any


class Transition:
    """
    A transition between two states.

    Parameters
    ----------
    from_state : State
        The state from which the transition starts.
    to_state : State
        The state to which the transition goes.
    condition : Any
        The condition under which the transition is taken.
    action : int
        The action to perform when the transition is taken.

    Attributes
    ----------
    from_state : State
        The state from which the transition starts.
    to_state : State
        The state to which the transition goes.
    condition : Any
        The condition under which the transition is taken.
    action : int
        The action to perform when the transition is taken.
    """

    def __init__(
        self,
        from_state: "State",
        to_state: "State",
        condition: Any = None,
        action: int = 1,
        negated: bool = False,
    ) -> None:
        self.from_state = from_state
        self.to_state = to_state
        self.condition = condition
        self.action = action
        self.negated = negated

    def check_condition(self, value: Any) -> bool:
        """
        Check if the transition is taken.

        Parameters
        ----------
        value : Any
            The value to check.

        Returns
        -------
        bool
            True if the transition is taken, False otherwise.
        """

        if self.condition is None:
            return True
        if isinstance(self.condition, list):
            ret_val = value in self.condition
        else:
            ret_val = value == self.condition
        ret_val = not ret_val if self.negated else ret_val
        neg = "^" if self.negated else ""
        logging.debug(
            f"Checking condition {neg}{self.str_cond} against {value} ({ret_val})"
        )
        return ret_val

    @property
    def is_epsilon(self) -> bool:
        """Returns True if the transition is an epsilon transition."""
        return self.condition is None

    def __str__(self) -> str:
        return f"{self.from_state} - ({self.condition}) -> {self.to_state}"

    @property
    def str_cond(self) -> str:
        """
        Get the string representation of the condition.

        Returns
        -------
        str
            The string representation of the condition.
        """

        if self.condition is None:
            return "ε"
        if isinstance(self.condition, list):
            cond = self.condition
            return (
                cond.__repr__()
                if len(cond) <= 10
                else (f"[{cond[0].__repr__()}, ..., {cond[-1].__repr__()}]")
            )
        return self.condition.__repr__()



from random import choice

from clismo.sim.optimizable_obj import OptimizableObject


class Step(OptimizableObject):
    def __init__(self, name, servers=None):
        super().__init__("servers")
        self.name = name
        self.servers = servers or []
        self.clients_queue = []
        self.attrs = {}

    def add_server(self, server):
        self.servers.append(server)

    def assign_client_to_server(self, client):
        free_servers = [server for server in self.servers if not server.in_use]
        rand_server = choice(free_servers)
        rand_server.in_use = True
        return rand_server.attend_client(client), rand_server

    def receive_client(self, client):
        if self.clients_queue or all(server.in_use for server in self.servers):
            self.clients_queue.append(client)
            return None, None
        return self.assign_client_to_server(client)

    def next_client(self):
        if self.clients_queue:
            client = self.clients_queue.pop(0)
            return self.assign_client_to_server(client)
        return None, None

    def __repr__(self):
        return self.name



"""
This module contains the structures for text tokenization.

Example of use:

    from toknizer import Tokenizer

    tknz = Tokenizer()
    tknz.add_pattern("AB", r"[ab]+")
    tknz.add_pattern("Int", r"[0-9]+", lambda lex: int(lex))

    text = "aaaba23bba34"

    tokens = tknz.tokenize(text)
    print(tokens)

    > [Tok('aaaba'), Tok(23), Tok('bba'), Tok(34)]
"""

from typing import Any, Callable, Dict, List, Union, Tuple

from clismo.exceptions import TokenizationError
from clismo.csre import RegexPattern, compile_patt


class Token:
    """Represents a single token.

    Parameters
    ----------
    token_type : str
        Token type name.
    lexem : str
        Token lexem.
    line : int
        Line token position.
    col : int
        Column token position.

    Attributes
    ----------
    token_type : str
        Token type name.
    lexem : str
        Token lexem.
    line : int
        Line token position.
    col : int
        Column token position.
    """

    def __init__(
        self, token_type: str, lexem: str, line: int = 0, col: int = 0
    ):
        self.token_type = token_type
        self.lexem = lexem
        self.line = line
        self.col = col

    def info(self) -> str:
        """Gives a detailed and formated info about the token.

        Returns
        -------
        str
            Token information.
        """
        line_col = f"(l:{self.line}, c:{self.col})"
        return f"{line_col:<16} {self.token_type:<25} {self.lexem}"

    def __getattr__(self, item):
        return self.token_type == item

    def __eq__(self, other):
        if not isinstance(other, Token):
            return self.lexem == other
        return super().__eq__(other)

    def __repr__(self):
        return f"Tok({self.lexem.__repr__()})"

    def __str__(self):
        return self.__repr__()


class Tokenizer:
    """Tokenizer"""

    def __init__(self):
        self.token_patterns: Dict[str, RegexPattern] = {}
        self._token_found_functions = {}
        self._process_tokens = lambda tk: tk
        self._keywords = {}

    def add_pattern(
        self, token_type: str, pattern: str, func: Callable[[str], Any] = None
    ):
        """Adds a pattern for recognizing tokens.

        Adding orden is very important as the patterns are checked in the same
        order they where added.

        Parameters
        ----------
        token_type : str
            Token type name for the pattern.
        pattern : str
            Regex pattern.
        func : Callable[[str], Any]
            When a match is found for a pattern and a token is created this
            function is applied to the token lexem. The token's new lexem will
            be the return value of `func`.
        """
        if token_type in self.token_patterns:
            raise TokenizationError(f"Token type {token_type} already exists.")
        self.token_patterns[token_type] = compile_patt(pattern)
        if func is None:
            func = lambda lex: lex
        self._token_found_functions[token_type] = func

    def add_keyword(self, keyword: str, token_type: str = None):
        """Adds a keyword to the tokenizer.

        Parameters
        ----------
        keyword : str
            Keyword to be added.
        token_type : str
            Token type name for the keyword.
        """
        if token_type is None:
            token_type = keyword
        self._keywords[keyword] = token_type

    def add_keywords(self, *keywords: Union[str, Tuple[str, str]]):
        """Adds a list of keywords to the tokenizer.

        Parameters
        ----------
        *keywords : Union[str, Tuple[str, str]]
            Keywords to be added.
        """
        for keyword in keywords:
            if isinstance(keyword, str):
                self.add_keyword(keyword)
            else:
                self.add_keyword(*keyword)

    def add_patterns(self, **kwargs: str):
        """Adds a list of patterns for recognizing tokens

        The order of patterns is very important as the patterns are checked in the same
        order they where added.

        Parameters
        ----------
        **kwargs: str


        Example
        -------
        tknz.add_patterns(
            INT = r"\\d+",
            OPERATOR = r"[+\\-*/]"
            )

        """
        for token_type, patt in kwargs.items():
            self.add_pattern(token_type, patt)

    def process_tokens(self, func: Callable[[List[Token]], List[Token]]):
        """Decorator for processing the tokens after tokenization process.

        The decorated function will recieve the raw token list resulting after
        tokenization and it should return the processed token list.

        Parameters
        ----------
        func : Callable[[List[Token]], List[Token]]
            Function to be decorated.
        """
        self._process_tokens = func
        return func

    def token(
        self, token_type: str, patt: str
    ) -> Callable[[Callable[[str], Any]], Callable[[str], Any]]:
        """Generates a decorator that adds a token pattern and use the decorated
        function to process the token lexem.

        Example
        -------

        This:

            @my_tokenizer.token("tok_type", r"re_patt")
            def tok_type(lexem: str) -> str:
                return lexem.lower()

        Is the same as:

            my_tokenizer.add_pattern(
                "tok_type", r"re_patt", lambda lexem: lexem.lower()
            )

        Parameters
        ----------
        token_type : str
            Token type name for the pattern.
        patt : str
            Regex pattern.

        Returns
        -------
        Callable[[Callable[[str], Any]], Callable[[str], Any]]
            Generated decorator.
        """

        def token_deco(func: Callable[[str], Any]) -> Callable[[str], any]:
            """Decorator that adds a token pattern and use the decorated
            function to process the token lexem.

            Parameters
            ----------
            func : Callable[[str], Any]
                Function that process the token lexem once the token is created.

            Returns
            -------
            Callable[[str], any]
                Decorated function.
            """
            self.add_pattern(token_type, patt, func)
            return func

        return token_deco

    def tokenize(self, text: str) -> List[Token]:
        """Tokenize a text using the added token patterns.

        Parameters
        ----------
        text : str
            Text to be tokenized.
        """
        tokens = []
        line, col = 0, 0
        i = 0
        while i < len(text):
            for token_type, patt in self.token_patterns.items():
                re_match = patt.match(text[i:])
                if re_match is not None:
                    lexem = re_match.matched_text
                    tok_lexem = self._token_found_functions[token_type](lexem)
                    if tok_lexem in self._keywords:
                        token_type = self._keywords[tok_lexem]
                    if tok_lexem is not None:
                        tok = Token(token_type, tok_lexem, line, col)
                        tokens.append(tok)
                    i += len(lexem)
                    line_breaks = lexem.count("\n")
                    line += line_breaks
                    col += len(lexem)
                    if line_breaks:
                        col = len(lexem.split("\n")[-1])
                    break
            else:
                raise TokenizationError(
                    f"No match found. Line: {line}, Col: {col}.\n"
                    f"Text: {text[i:i+10]}..."
                )

        return self._process_tokens(tokens)



"""
This module contains the TerminalSet structure.

This structure is used to calculate in an easy way the ``first`` and ``follow``
sets of a given grammar.
"""


from __future__ import annotations

from typing import Set

from clismo.compiler.grammar import Terminal
from clismo.compiler.tokenizer import Token


class TerminalSet:
    """Terminal set.

    Attributes
    ----------
    terminals : Set[Terminal]
        Set of terminals.
    """

    def __init__(self, terminals: Set[Terminal] = None):
        self.terminals = set() if terminals is None else terminals

    def __iter__(self):
        return iter(self.terminals)

    def add(self, terminal: Terminal):
        """Adds a terminal to the set.

        Parameters
        ----------
        terminal : Terminal
            terminal to add.
        """
        last_len = len(self.terminals)
        self.terminals.add(terminal)
        return last_len != len(self.terminals)

    def update(self, terminal_set: TerminalSet):
        """Updates the set with another set.

        Parameters
        ----------
        terminal_set : SymbolSet
            terminal set to update with.
        """
        last_len = len(self.terminals)
        self.terminals.update(terminal_set.terminals)
        return last_len != len(self.terminals)

    def __sub__(self, other):
        if isinstance(other, Terminal):
            return TerminalSet(terminals=self.terminals - {other})
        if isinstance(other, TerminalSet):
            return TerminalSet(terminals=self.terminals - other.terminals)
        if isinstance(other, set):
            return TerminalSet(terminals=self.terminals - other)
        if isinstance(other, str):
            new_set = {symbol for symbol in self.terminals if symbol.name != other}
            return TerminalSet(terminals=new_set)
        raise TypeError(
            f"unsupported operand type(s) for -: "
            f"'{type(self).__name__}' and '{type(other).__name__}'"
        )

    def __and__(self, other):
        if isinstance(other, Terminal):
            return TerminalSet(terminals=self.terminals & {other})
        if isinstance(other, TerminalSet):
            return TerminalSet(terminals=self.terminals & other.terminals)
        if isinstance(other, set):
            return TerminalSet(terminals=self.terminals & other)
        if isinstance(other, str):
            new_set = {symbol for symbol in self.terminals if symbol.name == other}
            return TerminalSet(terminals=new_set)
        raise TypeError(
            f"unsupported operand type(s) for &: "
            f"'{type(self).__name__}' and '{type(other).__name__}'"
        )

    def has_token(self, token: Token):
        """Checks if the set contains a token.

        Parameters
        ----------
        token : str
            token to check.

        Returns
        -------
        bool
            ``True`` if the set contains the token, ``False`` otherwise.
        """
        return any(terminal.name == token.token_type for terminal in self.terminals)

    def __len__(self):
        return len(self.terminals)

    def __contains__(self, symbol: Terminal):
        return symbol in self.terminals

    def __getitem__(self, index):
        return list(self.terminals)[index]

    def __repr__(self):
        return list(self.terminals).__repr__()



from typing import List

from numlab.compiler import Token, Tokenizer

tknz = Tokenizer()
tknz.add_pattern("NEWLINE", r"\n")
tknz.add_pattern("SPACE", r"( |\t)( |\t)*", lambda l: None)

tknz.add_pattern("COMMENT", r"#.*\n", lambda t: None)

# Augassing
tknz.add_pattern("+=", r"+=")
tknz.add_pattern("-=", r"-=")
tknz.add_pattern("*=", r"\*=")
tknz.add_pattern("@=", r"@=")
tknz.add_pattern("/=", r"/=")
tknz.add_pattern("%=", r"%=")
tknz.add_pattern("&=", r"&=")
tknz.add_pattern("|=", r"\|=")
tknz.add_pattern("^=", r"\^=")
tknz.add_pattern("<<=", r"<<=")
tknz.add_pattern(">>=", r">>=")
tknz.add_pattern("**=", r"\*\*=")
tknz.add_pattern("//=", r"//=")

# Operators
tknz.add_pattern("==", r"==")
tknz.add_pattern(">=", r">=")
tknz.add_pattern("<=", r"<=")
tknz.add_pattern("!=", r"!=")
tknz.add_pattern("**", r"\*\*")
tknz.add_pattern("<<", r"<<")
tknz.add_pattern(">>", r">>")
tknz.add_pattern("*", r"\*")
tknz.add_pattern("=", r"=")
tknz.add_pattern("|", r"\|")
tknz.add_pattern("^", r"\^")
tknz.add_pattern("&", r"&")
tknz.add_pattern("+", r"+")
tknz.add_pattern("-", r"-")
tknz.add_pattern("@", r"@")
tknz.add_pattern("/", r"/")
tknz.add_pattern("%", r"%")
tknz.add_pattern("//", r"//")
tknz.add_pattern("~", r"~")
tknz.add_pattern("<", r"<")
tknz.add_pattern(">", r">")

# Punctuation
tknz.add_pattern("(", r"\(")
tknz.add_pattern(")", r"\)")
tknz.add_pattern("{", r"{")
tknz.add_pattern("}", r"}")
tknz.add_pattern("[", r"[")
tknz.add_pattern("]", r"]")
tknz.add_pattern(";", r";")
tknz.add_pattern(",", r",")
tknz.add_pattern(".", r"\.")
tknz.add_pattern(":", r":")


# Keywords
tknz.add_keywords(
    "True",
    "False",
    "None",
    "and",
    "or",
    "not",
    "if",
    "else",
    "loop",
    "endloop",
    "nextloop",
    "return",
    "server",
    "client",
    "simulation",
    "step",
    "var",
    "from",
    "to",
    "by",
)

# Special terminals
tknz.add_pattern("NAME", r"(\a|\A|_)(\a|\A|\d|_)*")
tknz.add_pattern("NUMBER", r"\d\d*|\d\d*\.\d\d*")
tknz.add_pattern(
    "STRING", r"'((^')|(\\'))*(^\\)'|\"((^\")|(\\\"))*(^\\)\"", lambda t: t[1:-1]
)


@tknz.process_tokens
def process_tokens(tokens: List[Token]):
    indent_tok = Token("INDENT", "INDENT")
    dedent_tok = Token("DEDENT", "DEDENT")
    indentations = [0]
    new_tokens = []
    check_indent = 0
    for tok in tokens:
        if tok.NEWLINE and check_indent == 0:
            check_indent = 1
        if not tok.NEWLINE and check_indent == 1:
            check_indent = 2
        if check_indent == 2:
            new_indentation_size = tok.col
            while new_indentation_size < indentations[-1]:
                new_tokens.append(dedent_tok)
                indentations.pop()
            if new_indentation_size > indentations[-1]:
                indentations.append(new_indentation_size)
                new_tokens.append(indent_tok)
            check_indent = 0
        if not tok.NEWLINE or (new_tokens and not new_tokens[-1].NEWLINE):
            new_tokens.append(tok)

    if len(indentations) > 1:
        for _ in range(len(indentations) - 1):
            new_tokens.append(dedent_tok)

    new_tokens.append(Token("NEWLINE", "\n"))
    return new_tokens



"""
This module contains the structures for representing grammars.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import (Any, Callable, Dict, Iterator, List, Optional, Set, Tuple,
                    Union)

from clismo.compiler.generic_ast import AST
from clismo.compiler.tokenizer import Token, Tokenizer

# Tokenizer for grammars
TKNZ = Tokenizer()
TKNZ.add_pattern("NEWLINE", r"( |\n)*\n\n*( |\n)*", lambda l: "NEWLINE")
TKNZ.add_pattern("SPACE", r"( |\t)( |\t)*", lambda t: None)
TKNZ.add_pattern("COMMENT", r"#(^\n)*\n", lambda t: None)
TKNZ.add_pattern("LITERAL", r"'((^')|(\\'))*(^\\)'", lambda l: l[1:-1])
TKNZ.add_pattern("SPECIAL", r"EPS")
TKNZ.add_pattern("ID", r"(\a|\A|_)(\a|\A|\d|_)*")
TKNZ.add_pattern("OP", r"\||:")


@TKNZ.process_tokens
def _process_tokens(tokens: List[Token]) -> List[Token]:
    """Process the tokens of a grammar file.

    This allows grammars to have productions on independent lines.

    Parameters
    ----------
    tokens : List[Token]
        Grammar tokens.

    Returns
    -------
    List[Token]
        Processed tokens.
    """
    new_tokens = []
    skip = 0
    for i, tok in enumerate(tokens):
        if skip:
            skip -= 1
            continue
        next_tok = tokens[i + 1] if i < len(tokens) - 1 else None
        last_tok = tokens[i - 1] if i > 0 else None
        if not new_tokens and tok.NEWLINE:
            continue
        if (
            tok.NEWLINE
            and next_tok is not None
            and last_tok is not None
            and last_tok.OP
            and last_tok.lexem == ":"
            and next_tok.OP
            and next_tok.lexem == "|"
        ):
            skip = 1
            continue
        if (
            tok.NEWLINE
            and next_tok is not None
            and next_tok.OP
            and next_tok.lexem == "|"
        ):
            continue
        if tok.NEWLINE and new_tokens and new_tokens[-1].NEWLINE:
            continue

        new_tokens.append(tok)
    return new_tokens


class Symbol(metaclass=ABCMeta):
    """Abstract class for representing either a Terminal or a NonTerminal.

    Parameters
    ----------
    name :
        Name of the grammar symbol.

    Attributes
    ----------
    name :
        Name of the grammar symbol.
    """

    def __init__(self, name):
        self.name = name
        self._ast = None

    @property
    def ast(self):
        """Return the abstract syntax tree of the grammar symbol.

        Returns
        -------
        AST
            Abstract syntax tree of the grammar symbol.
        """
        if self.is_terminal:
            return self
        return self._ast

    @property
    def is_terminal(self) -> bool:
        """Checks if the grammar symbol is a terminal or not.

        Returns
        -------
        bool
            True if the grammar symbol is a Terminal.
        """
        return isinstance(self, Terminal)

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        return self is other

    @abstractmethod
    def copy(self) -> Symbol:
        """Creates a copy of the grammar symbol.

        Returns
        -------
        Symbol
            Copy of the grammar symbol.
        """
        pass

    def __hash__(self):
        return hash(self.name)


class Production:
    """Represents a grammar production.

    Parameters
    ----------
    symbols : List[Symbo]
        List of grammar symbols that compose the production.
    """

    def __init__(self, symbols: List[Symbol]):
        self.symbols = symbols
        self._head: NonTerminal = None
        self._builder: Callable = None

    def set_builder(self, func: Callable):
        """Sets the builder function for the production.

        Parameters
        ----------
        func : Callable
            Function that builds the AST.
        """
        self._builder = func

    @property
    def builder(self):
        """Returns the builder function for the production.

        Returns
        -------
        Callable
            Builder function.
        """
        if self._builder is None:
            raise ValueError("Builder function not set.")
        return self.set_builder

    @property
    def head(self):
        """Gets the head of the production.

        Returns
        -------
        NonTerminal
            Head of the production.
        """
        if self._head is None:
            raise ValueError("Production head missing.")
        return self._head

    @property
    def head_str(self) -> str:
        """Gets the head of the production as a string.

        Returns
        -------
        str
            Head of the production.
        """
        return f"{self.head.name} {str(self)}"

    @property
    def is_eps(self) -> bool:
        """Checks if the production is the empty production.

        Returns
        -------
        bool
            True if the production is the empty production.
        """
        return len(self.symbols) == 1 and self.symbols[0] == "EPS"

    def build_ast(self, symbols: List[Symbol]) -> AST:
        """Builds the AST for the production.

        Parameters
        ----------
        symbols : List[Symbol]
            List of grammar symbols that compose the production.

        Returns
        -------
        AST
            AST for the production.
        """
        if self._builder is None:
            raise ValueError(f"Builder function not set on production {self}.")
        return self._builder(*symbols)

    def __getitem__(self, index):
        return self.symbols[index]

    def __setitem__(self, index, item):
        self.symbols[index] = item

    def __repr__(self):
        prod_str = "-> " + " ".join(str(symbol) for symbol in self.symbols)
        return prod_str

    def __str__(self):
        return self.__repr__()


class NonTerminal(Symbol):
    """Represents a non terminal (a grammar expression).

    Parameters
    ----------
    name : str
        Name of the expression.
    prods : List[Production]
        List of productions that compose the expression.
    """

    def __init__(self, name, prods: Optional[List[Production]] = None):
        super().__init__(name)
        self.prods = [] if prods is None else prods
        for prod in self.prods:
            prod._head = self

    def set_ast(self, ast: AST):
        """Sets the abstract syntax tree for the expression.

        Parameters
        ----------
        ast : AST
            Abstract syntax tree for the expression.

        Returns
        -------
        None
        """
        self._ast = ast

    def copy(self):
        """Creates a copy of the non terminal.

        Returns
        -------
        NonTerminal
            Copy of the non terminal.
        """
        return NonTerminal(self.name, self.prods)

    def check_token(self, token: Token) -> bool:
        """Checks if the token matches the non terminal.

        Parameters
        ----------
        token : Token
            Token to be checked.

        Returns
        -------
        bool
            True if the token matches the non terminal.
        """
        return self.name == token.token_type

    def __getitem__(self, index):
        return self.prods[index]

    def __getattr__(self, item):
        if item.startswith("prod_"):
            return self.prods[int(item[5:])]
        raise AttributeError()

    def __repr__(self):
        return f"NT({self.name})"


class Terminal(Symbol):
    """Terminal.

    Parameters
    ----------
    name : str
        Name of the terminal.
    match :
        Regex pattern used to check if a token lexem match in parsing process.

    Attributes
    ----------
    name : str
        Name of the terminal.
    match : str
        Regex pattern used to check if a token lexem match in parsing process.
    """

    def __init__(self, name: str, value: Any = None):
        super().__init__(name)
        self.value = value

    def copy(self):
        """Creates a copy of the terminal.

        Returns
        -------
        Terminal
            Copy of the terminal.
        """
        return Terminal(self.name)

    def check_token(self, token: Token) -> bool:
        """Checks if the token matches the terminal.

        Parameters
        ----------
        token : Token
            Token to be checked.

        Returns
        -------
        bool
            True if the token matches the terminal.
        """
        return self.name == token.token_type

    def __repr__(self):
        return f"T({self.name})"


class Grammar:
    """Represents a Grammar.

    Parameters
    ----------
    exprs : List[NonTerminal]
        Grammar expressions.

    Attributes
    ----------
    exprs : List[NonTerminal]
        Grammar expressions.
    """

    def __init__(self, exprs: List[NonTerminal] = None):
        self.exprs = [] if exprs is None else exprs
        self.start = None
        if exprs:
            self.start = exprs[0]
        self.exprs_dict = {exp.name: exp for exp in exprs}

    def __getattr__(self, item):
        if item in self.exprs_dict:
            return self.exprs_dict[item]
        raise AttributeError()

    @property
    def symbols(self) -> List[Symbol]:
        """Returns the list of symbols.

        Returns
        -------
        List[Symbol]
            List of symbols.
        """
        non_term = [expr for expr in self.exprs]
        term = list(self.all_terminals())
        return non_term + term

    def assign_builders(self, builders: Dict[str, Callable]) -> None:
        """Assigns the builders for the productions.

        Parameters
        ----------
        builders : Dict[str, Callable]
            Dictionary of builders.
        """
        for expr in self.exprs:
            for prod in expr.prods:
                builder_name = f"{expr.name} {prod}"
                prod.set_builder(builders[builder_name])

    def add_expr(self, expr: NonTerminal):
        """Adds a grammar expression to the grammar.

        Parameters
        ----------
        expr : NonTerminal
            Grammar expression to be added.
        """
        if expr.name in self.exprs_dict:
            raise ValueError(f"Grammar expression {expr.name} already exists.")
        self.exprs.append(expr)
        self.exprs_dict[expr.name] = expr

    @property
    def start_expr(self):
        """Gets the start expression of the grammar.

        Returns
        -------
        NonTerminal
            Start expression.
        """
        if self.start is None:
            raise ValueError("Grammar has no start expression.")
        return self.start

    def all_terminals(self) -> Set[Terminal]:
        """Set of all the terminals in the grammar.

        Returns
        -------
        Set[Terminal]
            Terminals set.
        """
        terminals = []
        for exp in self.exprs_dict.values():
            for prod in exp.prods:
                for symbol in prod.symbols:
                    if isinstance(symbol, Terminal) and symbol.name not in [
                        "EPS",
                    ]:
                        terminals.append(symbol)
        return set(terminals)

    def all_productions(self) -> Iterator[Tuple[NonTerminal, Production]]:
        """Production iterator.

        In each iteration a tuple is given where the first element is the
        expression and the second is the production.

        Returns
        -------
        Iterator[Tuple[NonTerminal, Production]]
            (Expression, Production) iterator.
        """
        for expr in self.exprs:
            for prod in expr.prods:
                yield (expr, prod)

    @staticmethod
    def open(file_path: str) -> Grammar:
        """Reads and create a grammar from a `.gm` file.

        The `.grm` format can express a gramma as shown bellow.

        expression: production_1 | production_2 | ... | production_n

        Example:

        Expr: Term Expr_X
        Expr_X: '+' Expr | EPS
        Term: Factor Term_Y
        Term_Y: '*' Term | EPS
        Factor: '(' E ')' | i

        Productions also can be separated on independent lines. The follow
        grammar is the same as the show as a example:

        Expr: Term Expr_X

        Expr_X:
            | '+' Expr
            | EPS

        Term: Factor Term_Y

        Term_Y:
            | '*' Term
            | EPS

        Factor:
            | '(' E ')'
            | i

        Parameters
        ----------
        file_path : str
            Grammar file path.

        Returns
        -------
        Grammar
            Readed grammar.
        """
        with open(file_path, "r", encoding="utf-8") as file:
            text = file.read()
        tokens = TKNZ.tokenize(text)
        grm_parser = _GrammarParser(tokens)
        grm = grm_parser.parse()
        return grm


class _GrammarParser:
    """Parser for grammars.

    Parameters
    ----------
    tokens : List[Token]
        Tokens to be parsed.
    """

    def __init__(self, tokens: List[Token]):
        self.tokens = tokens
        self._cursor = 0

    def _check_token(
        self,
        expected_type: Union[str, List[str]],
        expected_value: str = None,
        pos: int = None,
        or_none: bool = False,
    ):
        """Check if the token in a given position match the expected type and
        value.

        Parameters
        ----------
        expected_type : Union[str, List[str]]
            Expected token type.
        expected_value : str, optional
            Expected value (lexem) of the token. If None then the value is not
            checked, by default None.
        pos : int, optional
            Position of the token to be checked. If None then `pos` is taken as
            the cursor prosition, by default None.
        or_none : bool, optional
            Do not raise exception if the position is invalid.

        Raises
        ------
        ValueError
            If the token does not match with the given restrictions.
        """
        if pos is None:
            pos = self._cursor
        if not isinstance(expected_type, list):
            expected_type = [expected_type]
        if pos >= len(self.tokens):
            if not or_none:
                raise ValueError("Expected token type: {expected_type}")
        if self.tokens[pos].token_type not in expected_type:
            raise ValueError(
                f"Error parsing token {self.tokens[pos]}. Expected "
                f"token type: {expected_type}."
            )
        if expected_value and self.tokens[pos].lexem != expected_value:
            raise ValueError(
                f"Error parsing token {self.tokens[pos]}. Expected "
                f"token value: {expected_value}"
            )

    @property
    def _ctoken(self) -> Token:
        """Token as cursos position (current token)self."""
        if self._cursor >= len(self.tokens):
            return None
        return self.tokens[self._cursor]

    def parse(self) -> Grammar:
        """Parses the tokens.

        Returns
        -------
        Grammar
            Resulting grammar.
        """
        self._cursor = 0
        return self._parse_grammar()

    def _parse_grammar(self) -> Grammar:
        """Parse a grammar.

        Returns
        -------
        Grammar
            Resulting grammar.
        """
        if self._cursor >= len(self.tokens):
            return None
        self._check_token("ID")
        exprs = self._parse_expr_list()
        exprs_dict = {exp.name: exp for exp in exprs}
        terminals = {}
        for exp in exprs:
            for prod in exp.prods:
                for i, symbol in enumerate(prod.symbols):
                    if symbol.name in exprs_dict:
                        prod[i] = exprs_dict[symbol.name]
                    elif symbol.name in terminals:
                        prod[i] = terminals[symbol.name]
                    else:
                        terminals[symbol.name] = symbol
        return Grammar(exprs)

    def _parse_expr_list(self) -> List[NonTerminal]:
        """Parses an expression list.

        Returns
        -------
        List[NonTerminal]
            Resulting expressions.
        """
        if self._ctoken is None:
            return None
        expr = self._parse_expr()
        expr_list = self._parse_expr_list()
        exprs = [expr]
        if expr_list is not None:
            exprs += expr_list
        return exprs

    def _parse_expr(self) -> NonTerminal:
        """Parses an expression.

        Returns
        -------
        NonTerminal
            Resulting expresson.
        """
        self._check_token("ID")
        expr_name = self._ctoken.lexem
        self._cursor += 1
        self._check_token("OP", ":")
        self._cursor += 1
        prod = self._parse_prod()
        prod_list = self._parse_prod_list()
        self._check_token("NEWLINE", or_none=True)
        self._cursor += 1
        prods = [prod]
        if prod_list is not None:
            prods += prod_list
        expr = NonTerminal(expr_name, prods)
        return expr

    def _parse_prod(self) -> Production:
        """Parses a production.

        Returns
        -------
        Production
            Resulting production.
        """
        prod_symbols = self._parse_prod_symbols()
        return Production(prod_symbols)

    def _parse_prod_symbols(self) -> List[Symbol]:
        """Parses all the symbols of a production.

        Returns
        -------
        List[Symbol]
            Resulting symbols.
        """
        if self._ctoken.OP:
            return None
        if self._ctoken.NEWLINE:
            return None
        symbol = self._parse_symbol()
        prod_symbols = self._parse_prod_symbols()
        if prod_symbols is None:
            return [symbol]
        return [symbol] + prod_symbols

    def _parse_symbol(self) -> Symbol:
        """Parses a grammar symbol.

        Returns
        -------
        Symbol
            Resulting symbol.
        """
        self._check_token(["ID", "LITERAL", "SPECIAL"])
        name = self._ctoken.lexem

        # It always returns a Terminal but after parsing all terminals with
        # non terminal names will be replaced by the non terminal symbols.
        term = Terminal(name)
        self._cursor += 1
        return term

    def _parse_prod_list(self) -> List[Production]:
        """Parses a production list.

        Returns
        -------
        List[Production]
            Resulting productions.
        """
        if self._cursor >= len(self.tokens):
            return None
        if self._ctoken.NEWLINE:
            return None
        self._check_token("OP", "|")
        self._cursor += 1
        prod = self._parse_prod()
        prod_list = self._parse_prod_list()
        if prod_list is None:
            return [prod]
        return [prod] + prod_list



from __future__ import annotations

from typing import Any, Dict, Optional


class Context:
    def __init__(self, parent: Context = None):
        self.parent: Context = parent
        self.symbols: Dict[str, Any] = {}
        self.var_count: int = 0

    def make_child(self) -> Context:
        return Context(self)

    def define(self, name: str, value: Any):
        self.symbols[name] = value

    def resolve(self, name: str) -> Optional[Any]:
        if name in self.symbols:
            return self.symbols[name]
        if self.parent:
            return self.parent.resolve(name)
        return None



from queue import PriorityQueue
from typing import Callable, List


class Server:
    def __init__(self, func: Callable, cost: float = 1.0):
        self.func = func
        self.cost = cost


class ClientServerModel:
    def __init__(
        self,
        arrival_func: Callable,
        servers: List[Server],
        config: List[int],
        client_limit: int = None,
        time_limit: float = None,
    ):
        self.arrival_func = arrival_func
        if len(config) != len(servers):
            raise ValueError("config and servers must have same length")
        for conf in config:
            if conf <= 0:
                raise ValueError("config values must be positive")
        self.servers = servers
        self.config = config
        self.events = PriorityQueue()
        self.servers_queue = [0] * len(servers)
        self.servers_in_use = [0] * len(servers)
        self.time = 0.0
        self.clients = 0
        if client_limit is None and time_limit is None:
            raise ValueError("Either client_limit or time_limit must be specified")
        self.client_limit = client_limit
        self.time_limit = time_limit

    def run(self):
        self.time = 0
        self.clients = 0
        self.servers_queue = [0] * len(self.servers)
        self.servers_in_use = [0] * len(self.servers)
        while True:
            if self.events.empty():
                self.events.put((self.time + self.arrival_func(), 0))
            time, server = self.events.get()
            if self.time_limit is not None and time > self.time_limit:
                break
            self.time = time
            if server > 0:
                if self.servers_queue[server - 1]:
                    self.servers_queue[server - 1] -= 1
                    self.events.put(
                        (self.time + self.servers[server - 1].func(), server)
                    )
                else:
                    self.servers_in_use[server - 1] -= 1
            if server == len(self.servers):
                self.clients += 1
                if self.client_limit is not None and self.clients > self.client_limit:
                    break
                continue
            if server == 0:
                self.events.put((self.time + self.arrival_func(), 0))

            if self.servers_in_use[server] >= self.config[server]:
                self.servers_queue[server] += 1
            else:
                self.servers_in_use[server] += 1
                self.events.put((self.time + self.servers[server].func(), server + 1))



from pathlib import Path

import typer

from clismo.compiler import Grammar, LR1Parser, ParserManager
from clismo.cs_builders import builders
from clismo.cs_tokenizer import tknz
from clismo.visitors.eval_visitor import EvalVisitor
from clismo.visitors.semantic_checker import SemanticChecker
from clismo.visitors.type_builder import TypeBuilder

app = typer.Typer(add_completion=False)


def echo(msg, verbose):
    if verbose:
        typer.echo(msg)


def get_ast(file_path: str, verbose: bool = False):
    # Load grammar
    echo("Loading grammar", verbose)
    grammar = Grammar.open(str(Path(__file__).parent / "cs_grammar.gm"))
    echo("Assigning builders", verbose)
    grammar.assign_builders(builders)

    # Create LR1Parser
    echo("Loading parser table", verbose)
    parser = LR1Parser(grammar, str(Path(__file__).parent / "cs_lr1_table"))

    # Create parser
    parser_man = ParserManager(grammar, tknz, parser)

    # Parse file
    echo("Parsing script", verbose)
    program = parser_man.parse_file(file_path)
    return program


@app.command()
def run(
    input_path: str = typer.Argument(..., help="Input file"),
    dump: bool = typer.Option(False, "--dump", "-d", help="Dump AST"),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Verbose mode"),
):
    program = get_ast(input_path, verbose)
    if dump:
        program.dump()

    type_builder = TypeBuilder()
    type_builder.visit(program)

    semantic_checker = SemanticChecker(type_builder.types)
    semantic_checker.visit(program)

    evaluator = EvalVisitor(semantic_checker.objects)
    evaluator.visit(program)


if __name__ == "__main__":
    app()



import operator
import random

import clismo.builtin as builtin
import clismo.cs_ast as ast
from clismo.lang.type import Type
from clismo.lang.visitor import Visitor
from clismo.optimization.client_server_optimizer import ModelOptimizer
from clismo.sim.client import Client
from clismo.sim.server import Server
from clismo.sim.simulation import Simulation
from clismo.sim.step import Step

# pylint: disable=function-redefined
# pylint: disable=missing-function-docstring
# pylint: disable=unpacking-non-sequence


OPERATOR_FUNC = {
    ast.Operator.ADD: operator.add,
    ast.Operator.SUB: operator.sub,
    ast.Operator.MUL: operator.mul,
    ast.Operator.DIV: operator.truediv,
    ast.Operator.POW: operator.pow,
    ast.Operator.MOD: operator.mod,
    ast.Operator.POW: operator.pow,
    ast.Operator.LSHIFT: operator.lshift,
    ast.Operator.RSHIFT: operator.rshift,
    ast.Operator.BIT_XOR: operator.xor,
    ast.Operator.BIT_AND: operator.and_,
    ast.Operator.BIT_OR: operator.or_,
    ast.Operator.FLOORDIV: operator.floordiv,
    ast.Operator.EQ: operator.eq,
    ast.Operator.NOT_EQ: operator.ne,
    ast.Operator.LT: operator.lt,
    ast.Operator.GT: operator.gt,
    ast.Operator.LTE: operator.le,
    ast.Operator.GTE: operator.ge,
    ast.Operator.UADD: lambda x: +x,
    ast.Operator.USUB: lambda x: -x,
    ast.Operator.INVERT: operator.invert,
    ast.Operator.NOT: operator.not_,
}


def get_real_value(obj):
    val = obj.value
    if (
        isinstance(val, list)
        and val
        and not isinstance(val[0], (Client, Server, Step, Simulation))
    ):
        val = [item.value for item in val]
    return val


def get_clismo_type(val):
    if isinstance(val, bool):
        return builtin.cs_bool(val)
    if isinstance(val, int):
        return builtin.cs_int(val)
    if isinstance(val, float):
        return builtin.cs_float(val)
    if isinstance(val, str):
        return builtin.cs_str(val)
    if isinstance(val, list):
        items = [get_clismo_type(item) for item in val]
        if isinstance(items[0], (Client, Server, Step, Simulation)):
            type_ = type(items[0]).__name__.lower()
        else:
            type_ = items[0].type.type_name
        return Type.get(f"{type_}_list")(items)
    if isinstance(val, (Client, Server, Step, Simulation)):
        return val
    raise TypeError(f"Can not convert type {type(val)}")


class EvalVisitor:
    visitor = Visitor().visitor

    def __init__(self, objects):
        self.simulation = None
        self.current_obj = None
        self.current_client = None
        self.objects = objects
        self.objects_by_name = {}
        for obj in self.objects.values():
            self.objects_by_name.update(obj)
        self.context = {}
        self.flags = {
            "return": None,
            "nextloop": False,
            "endloop": False,
        }

    def resolve(self, name):
        if name == "self":
            return self.current_obj
        if name == "time":
            return builtin.cs_float(self.simulation.time)
        if name == "clients":
            return builtin.cs_int(self.simulation.clients)
        if name == "current_client":
            return self.current_client
        if name in self.context:
            return self.context[name]
        if name in self.objects_by_name:
            return self.objects_by_name[name]
        raise Exception(f"Unknown name: {name}")

    @visitor
    def visit(self, node: ast.Program):
        for stmt in node.stmts:
            self.visit(stmt)
        for sim in self.objects["Simulations"].values():
            self.simulation = sim
            mode = sim.attrs.get("mode", "run")
            max_iter = sim.attrs.get("max_iter", 5)
            pop_size = sim.attrs.get("pop_size", 10)
            mut_prob = sim.attrs.get("mut_prob", 0.1)
            best_sel = sim.attrs.get("best_sel", 3)
            new_rand = sim.attrs.get("new_rand", 2)
            logs = sim.attrs.get("logs", False)

            if mode == "run":
                sim.run(logs)
                print("Total time:", sim.time)
                print("Clients attended:", sim.clients)
            else:
                opt = ModelOptimizer(
                    sim,
                    max_iter=max_iter,
                    population_size=pop_size,
                    mutation_prob=mut_prob,
                    best_selection_count=best_sel,
                    generate_new_randoms=new_rand,
                )
                opt.run()

    @visitor
    def visit(self, node: ast.ClientDef):
        self.current_obj = self.objects_by_name[node.name]
        for item in node.body:
            self.visit(item)

    @visitor
    def visit(self, node: ast.ServerDef):
        self.current_obj = self.objects_by_name[node.name]
        for item in node.body:
            self.visit(item)

    @visitor
    def visit(self, node: ast.StepDef):
        self.current_obj = self.objects_by_name[node.name]
        for item in node.body:
            self.visit(item)

    @visitor
    def visit(self, node: ast.SimulationDef):
        self.current_obj = self.objects_by_name[node.name]
        for item in node.body:
            self.visit(item)

    @visitor
    def visit(self, node: ast.Attr):
        obj = self.current_obj
        val = get_real_value(self.visit(node.value))
        if isinstance(obj, Step):
            if node.name == "servers":
                obj.servers = val
            else:
                obj.attrs[node.name] = val
        if isinstance(obj, Simulation):
            if node.name == "time_limit":
                obj.time_limit = val
            if node.name == "client_limit":
                obj.client_limit = val
            if node.name == "steps":
                obj.steps = val
            else:
                obj.attrs[node.name] = val
        else:
            obj.attrs[node.name] = val

    @visitor
    def visit(self, node: ast.Function):
        obj = self.current_obj

        def func():
            self.current_obj = obj
            for stmt in node.body:
                self.visit(stmt)
            val = self.flags.get("return", None)
            self.flags["return"] = None
            return get_real_value(val)

        if node.name == "possible":
            attr_name = node.info[0]
            obj.add_possible_change(func, attr_name)

        if node.name == "on_server_out":
            for server in node.info:
                name = server
                obj.add_on_server_out_callback(func, name)

        if node.name == "attend_client":

            def attend_client(server, client):
                self.current_obj = server
                self.current_client = client
                for stmt in node.body:
                    self.visit(stmt)
                val = self.flags.get("return", None)
                self.flags["return"] = None
                return get_real_value(val)

            self.current_obj.func = attend_client

        if node.name == "arrive":
            for client in node.info:
                obj.add_arrival_func(func, self.objects["Clients"][client])

        if node.name == "minimize":
            obj.minimize_func = func

    @visitor
    def visit(self, node: ast.Assign):
        val = self.visit(node.value)
        self.context[node.name] = val

    @visitor
    def visit(self, node: ast.If):
        cond_val = get_real_value(self.visit(node.cond))
        if cond_val:
            for stmt in node.then:
                self.visit(stmt)
                if (
                    self.flags["endloop"]
                    or self.flags["nextloop"]
                    or self.flags["return"] is not None
                ):
                    break
        elif node.els:
            for stmt in node.els:
                self.visit(stmt)
                if (
                    self.flags["endloop"]
                    or self.flags["nextloop"]
                    or self.flags["return"] is not None
                ):
                    break

    @visitor
    def visit(self, node: ast.Return):
        val = self.visit(node.value)
        self.flags["return"] = val
        return val

    @visitor
    def visit(self, node: ast.Loop):
        start = builtin.cs_int(0) if node.start is None else self.visit(node.start)
        end = builtin.cs_int(-1) if node.end is None else self.visit(node.end)
        step = builtin.cs_int(1) if node.step is None else self.visit(node.step)
        while start.value < end.value:
            if node.target:
                self.context[node.target] = start
            for stmt in node.body:
                self.visit(stmt)
                if self.flags["return"] is not None:
                    break
                if self.flags["nextloop"]:
                    self.flags["nextloop"] = False
                    continue
                if self.flags["endloop"]:
                    self.flags["endloop"] = False
                    break
            start.value += step.value
        if node.target:
            self.context.pop(node.target)

    @visitor
    def visit(self, node: ast.EndLoop):
        self.flags["endloop"] = True

    @visitor
    def visit(self, node: ast.NextLoop):
        self.flags["nextloop"] = True

    @visitor
    def visit(self, node: ast.Call):
        args = [self.visit(arg) for arg in node.args]
        if node.name == "get":
            obj, val = args[0], get_real_value(args[1])
            if not isinstance(obj, (Client, Server, Step, Simulation)):
                raise ValueError(
                    "get() can only be called on clients, servers, steps or simulations"
                )
            if hasattr(obj, val):
                return get_clismo_type(getattr(obj, get_real_value(args[1])))
            if val in obj.attrs:
                return get_clismo_type(obj.attrs[val])
            raise Exception(f"{obj.name} has no attribute {val}")
        if node.name == "set":
            obj, attr, val = args[0], get_real_value(args[1]), get_real_value(args[2])
            if not isinstance(obj, (Client, Server, Step, Simulation)):
                raise ValueError(
                    "set() can only be called on clients, servers, steps or simulations"
                )
            if hasattr(obj, attr):
                setattr(obj, attr, val)
            else:
                obj.attrs[attr] = val
            return builtin.cs_none(None)
        return builtin.resolve(node.name)[0](*args)

    @visitor
    def visit(self, node: ast.BinOp):
        oper = OPERATOR_FUNC[node.op]
        left_val = get_real_value(self.visit(node.left))
        right_val = get_real_value(self.visit(node.right))
        return get_clismo_type(oper(left_val, right_val))

    @visitor
    def visit(self, node: ast.UnaryOp):
        oper = OPERATOR_FUNC[node.op]
        val = self.visit(node.expr)
        return get_clismo_type(oper(get_real_value(val)))

    @visitor
    def visit(self, node: ast.ListExpr):
        elements = [self.visit(item) for item in node.elements]
        if isinstance(elements[0], (Client, Server, Step, Simulation)):
            type_ = type(elements[0]).__name__.lower()
        else:
            type_ = elements[0].type.type_name
        return Type.get(f"{type_}_list")(elements)

    @visitor
    def visit(self, node: ast.Name):
        return self.resolve(node.name)

    @visitor
    def visit(self, node: ast.Constant):
        val = node.value
        if isinstance(val, bool):
            return builtin.cs_bool(val)
        if isinstance(val, int):
            return builtin.cs_int(val)
        if isinstance(val, float):
            return builtin.cs_float(val)
        if isinstance(val, str):
            return builtin.cs_str(val)



import clismo.cs_ast as ast
from clismo.lang.visitor import Visitor

# pylint: disable=function-redefined
# pylint: disable=missing-function-docstring
# pylint: disable=unpacking-non-sequence


class TypeBuilder:
    visitor = Visitor().visitor

    def __init__(self):
        self.types = {
            "Clients": {
                "DefaultClient": {},
            },
            "Servers": {},
            "Steps": {},
            "Simulations": {},
        }

    @visitor
    def visit(self, node: ast.Program):
        for stmt in node.stmts:
            self.visit(stmt)

    def _add_type(self, type_, name, attrs):
        self.types[type_][name] = {}
        for attr in attrs:
            key, val = self.visit(attr)
            if isinstance(val, tuple):
                overloads = self.types[type_][name].get(key, {})
                overloads[val[0]] = val[1]
                self.types[type_][name][key] = overloads
            else:
                self.types[type_][name][key] = val

    @visitor
    def visit(self, node: ast.ClientDef):
        self._add_type("Clients", node.name, node.body)

    @visitor
    def visit(self, node: ast.ServerDef):
        self._add_type("Servers", node.name, node.body)

    @visitor
    def visit(self, node: ast.StepDef):
        self._add_type("Steps", node.name, node.body)

    @visitor
    def visit(self, node: ast.SimulationDef):
        self._add_type("Simulations", node.name, node.body)

    @visitor
    def visit(self, node: ast.Attr):
        return node.name, node.value

    @visitor
    def visit(self, node: ast.Function):
        return node.name, (tuple(node.info), node.body)



from clismo.sim.optimizable_obj import OptimizableObject


class Server(OptimizableObject):
    def __init__(self, name, func, **attrs):
        super().__init__()
        self.name = name
        self.func = func
        self.attrs = attrs
        self.in_use = False

    def attend_client(self, client):
        self.in_use = True
        return self.func(self, client)

    @staticmethod
    def ghost():
        return Server("ghost", lambda s, c: None)

    def __lt__(self, other):
        return self.name < other.name

    def get(self):
        return Server(self.name, self.func, **self.attrs)

    def __repr__(self):
        return self.name



import logging
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Set, Tuple, Union

from clismo.compiler.generic_ast import AST
from clismo.compiler.grammar import (Grammar, NonTerminal, Production, Symbol,
                                     Terminal)
from clismo.compiler.grammar_ops import calculate_first, calculate_follow
from clismo.compiler.parsers.lritem import LRItem
from clismo.compiler.parsers.parser import Parser
from clismo.compiler.tokenizer import Token
from clismo.exceptions import ParsingError


class LR1Table:
    """
    This class represents the LR1 table.
    """

    def __init__(self, grammar: Grammar, table_file: str = None):
        """
        Initializes a new LR1 table.

        Parameters
        ----------
        grammar : Grammar
            Grammar that will be used.
        table_file : str, optional
            File to load the table from.

            If file does not the table will be created and saved to this file.
            If not given, the table will be generated.
        """
        self.grammar = grammar
        self._prepare_grammar()
        self._symbols = {sym.name: sym for sym in grammar.symbols}
        self._symbols["$"] = Terminal("$")
        self._productions = {
            prod.head_str: prod for _, prod in grammar.all_productions()
        }

        # [state, symbol] -> [next_state, production]
        self._table: Dict[Tuple[int, str], Union[str, int, Production]] = {}

        self._first = None
        self._follow = None
        self._states_by_id: Dict[int, List[LRItem]] = None
        self._id_by_hashsum: Dict[int, int] = {}
        self._closure_cache: Dict[int, List[LRItem]] = {}
        self._goto_cache: Dict[Tuple[int, str], int] = {}
        self._lr_items: Dict[Production, int, Terminal] = None
        self._item_prods: Dict[NonTerminal, Production] = None

        if table_file is not None:
            table_file_path = Path(table_file)
            if table_file_path.exists():
                self._load_table(table_file)
                return
        self._build_table(table_file)

    def _load_table(self, table_file: str):
        """Loads the table from a file.

        Parameters
        ----------
        table_file : str
            Path to the file.
        """
        logging.info(f"Loading table from {table_file}")
        with open(str(table_file), "r", encoding="utf-8") as table_f:
            file_lines = table_f.readlines()

        assert len(file_lines) % 3 == 0, "Invalid table file"

        for i in range(0, len(file_lines), 3):
            state = int(file_lines[i])
            symbol = self._symbols[file_lines[i + 1].strip()]
            str_t_val = file_lines[i + 2].strip()
            t_val = str_t_val
            if "->" in str_t_val:
                is_eps = False
                if str_t_val.endswith("->"):
                    is_eps = True
                    str_t_val += " EPS"
                t_val = self._productions[str_t_val]
                if is_eps:
                    item_prod = Production([])
                    item_prod._head = t_val.head
                    item_prod._builder = t_val._builder
                    t_val = item_prod
            elif str_t_val.isnumeric():
                t_val = int(str_t_val)
            self._table[(state, symbol)] = t_val

    def save_table(self, table_file: str):
        """Saves the table to a file.

        Parameters
        ----------
        table_file : str
            Path to the file.
        """
        logging.info(f"Saving table to {table_file}")
        with open(table_file, "w", encoding="utf-8") as table_f:
            for key, value in self._table.items():
                state, symbol = key
                t_val = "" if value is None else str(value)
                if isinstance(value, Production):
                    t_val = value.head_str
                table_f.write(f"{state}\n")
                table_f.write(f"{symbol}\n")
                table_f.write(f"{t_val}\n")

    def _prepare_grammar(self):
        logging.info("Preparing grammar (adding S')")
        if "S`" not in self.grammar.exprs_dict:
            non_ter_prod = Production([self.grammar.start])
            non_ter = NonTerminal("S`", [non_ter_prod])
            self.grammar.add_expr(non_ter)
            self.grammar.start = non_ter
            self.grammar.start.prod_0.set_builder(lambda s: s.ast)

    def _extract_grammar_lr_items(self) -> List[LRItem]:
        """Extracts all LR items from the grammar."""

        logging.info("Extracting all LR items")
        lr_items = []
        self._item_prods = {}
        for _, prod in self.grammar.all_productions():
            for dot_pos in range(len(prod.symbols) + 1):
                item_prod = prod
                is_eps = prod.is_eps
                if is_eps:
                    item_prod = Production([])
                    item_prod._head = prod.head
                    item_prod._builder = prod._builder
                slr_item = LRItem(item_prod, dot_pos)
                lr_items.append(slr_item)
                if prod.head not in self._item_prods:
                    self._item_prods[prod.head] = []
                self._item_prods[prod.head.name].append(item_prod)
                if is_eps:
                    break
        logging.info(f"Found {len(lr_items)} LR items")
        logging.debug("Exacted items in the LR automata:")
        for lr_item in lr_items:
            logging.debug(f"  {lr_item}")
        return lr_items

    def _contained_in_first(self, terminal: Terminal, *symbols):
        """Checks if a terminal is contained in the first set of a set of symbols.

        Parameters
        ----------
        terminal : Terminal
            Terminal to be checked.
        symbols : Symbol
            Symbols to be checked.

        Returns
        -------
        bool
            True if the terminal is contained in the first set of the symbols.
        """
        for symbol in symbols:
            if isinstance(symbol, Symbol) and symbol.is_terminal:
                return symbol == terminal
            if terminal in self._first[symbol]:
                return True
            if "EPS" not in self._first[symbol]:
                break
        return False

    def _build_table(self, table_file: str = None):
        self._first = calculate_first(self.grammar)
        self._follow = calculate_follow(self.grammar, self._first)
        items = self._extract_grammar_lr_items()

        self._lr_items = {}
        for item in items:
            for follow in self._follow[item.prod.head]:
                new_lr_item = LRItem(item.prod, item.dot_pos, follow)
                self._lr_items[item.prod, item.dot_pos, follow] = new_lr_item

        init_state = self._closure(
            {
                self._lr_items[
                    self.grammar.start_expr.prod_0,
                    0,
                    self._follow[self.grammar.start_expr][0],
                ]
            }
        )

        self._states_by_id = {0: init_state}

        logging.info("Building LR1 table")
        lr1_table: Dict[Tuple[int, str], Union[str, int, Production]] = {}
        current_state = 0
        while current_state < len(self._states_by_id):
            logging.info(f"Building state {current_state} of {len(self._states_by_id)}")
            state = self._states_by_id[current_state]
            for item in state:
                if item.at_symbol is None:
                    val = "OK" if item.prod.head.name == "S`" else item.prod
                    table_key = (current_state, item.lah.name)
                else:
                    val = self._goto(current_state, item.at_symbol)
                    table_key = (current_state, item.at_symbol.name)

                cont_val = lr1_table.get(table_key, None)

                if cont_val is not None and cont_val != val:
                    raise ValueError(
                        f"LR1 table already contains "
                        f"{table_key} -> {cont_val.__repr__()}  *** {val.__repr__()}"
                    )
                lr1_table[table_key] = val
            current_state += 1
        self._table = lr1_table
        logging.info("LR1 table built")
        if table_file is not None:
            self.save_table(table_file)

    def get_state_number(self, items: Set[LRItem]) -> int:
        """Returns the state number for a list of LR items.

        Parameters
        ----------
        items : List[LRItem]
            List of LR items.

        Returns
        -------
        int
            State number.
        """
        hashsum = self.items_hash(items)
        if hashsum in self._id_by_hashsum:
            return self._id_by_hashsum[hashsum]
        number = len(self._states_by_id)
        self._states_by_id[number] = items
        self._id_by_hashsum[hashsum] = number
        return number

    def __getitem__(self, index):
        return self._table.get(index, None)

    def items_hash(self, items: Set[LRItem]) -> Any:
        """Returns a unique value for a list of LR items.

        Parameters
        ----------
        items : Set[LRItem]
            Set of LR items.

        Returns
        -------
        Any
            Hash sum.
        """
        # return "".join([item.__repr__() for item in items])
        return sum(hash(item) for item in items)

    @lru_cache(maxsize=None)
    def _goto(self, state: int, symbol: Symbol) -> int:
        """Returns the state number for a state and a symbol.

        Parameters
        ----------
        state : int
            State number.
        symbol : Symbol
            Symbol.

        Returns
        -------
        int
            State number.
        """
        logging.debug(f"Goto({state}, {symbol})")
        state_items = self._states_by_id[state]
        filtered_items = {
            self._lr_items[item.prod, item.dot_pos + 1, item.lah]
            for item in state_items
            if item.at_symbol == symbol
        }
        clausure = self._closure(filtered_items)
        return self.get_state_number(clausure)

    def _closure(self, items: Set[LRItem]) -> Set[LRItem]:
        """Returns the closure of a list of LR items.

        Parameters
        ----------
        items : List[LRItem]
            List of LR items.

        Returns
        -------
        List[LRItem]
            Closure of the list of LR items.
        """
        hashsum = self.items_hash(items)
        if hashsum in self._closure_cache:
            return self._closure_cache[hashsum]
        logging.debug(f"Calculating closure of {items}")
        closure = items
        change = True
        while change:
            change = False
            new = set()
            for item in closure:
                next_item = item.at_symbol
                if next_item is None or next_item.is_terminal:
                    continue
                lah = item.lah
                rest = item.prod.symbols[item.dot_pos + 1 :]
                rest.append(lah)
                for prod in self._item_prods[next_item.name]:
                    for fol in self._follow[next_item]:
                        if not self._contained_in_first(fol, *rest):
                            continue
                        lr_item = self._lr_items[prod, 0, fol]
                        if lr_item not in closure:
                            new.add(lr_item)
                            change = True
            closure.update(new)
        self._closure_cache[hashsum] = closure
        return closure


class LR1Parser(Parser):
    """LR1 Parser.

    Parameters
    ----------
    grammar : Grammar
        Grammar to be used.
    table_file : str
        Path to the file containing the LR1 table.

        If the file does not exist, it will be created.
        If not specified, the table will be built automatically from
        the grammar.
    """

    def __init__(self, grammar: Grammar, table_file: str = None):
        super().__init__(grammar)
        self.lr1_table = LR1Table(grammar, table_file)

    def save_table(self, table_file: str):
        """Saves the LR1 table."""
        self.lr1_table.save_table(table_file)

    def parse(self, tokens: List[Token]) -> AST:
        logging.debug(f"Parsing {len(tokens)} tokens (LR1)")
        table = self.lr1_table
        stack: List[Tuple[Symbol, int]] = []
        i = 0
        while i < len(tokens):
            token = tokens[i]
            logging.info(f"----------------------------------------------------")
            logging.info(f"Parsing token {token}. Stack: {stack}")
            current_state = stack[-1][1] if stack else 0
            table_val = table[current_state, token.token_type]
            logging.info(
                f"Table value: {table_val} at ({current_state}, {token.token_type})"
            )
            if table_val == "OK":
                break
            if isinstance(table_val, int):
                logging.info(f"Making SHIFT action")
                term = Terminal(token.token_type, value=token.lexem)
                stack.append((term, table_val))
                i += 1
            elif isinstance(table_val, Production):
                logging.info(f"Making REDUCE action")
                reduce_prod = table_val
                # Pop from stack the necessary items
                items_needed = len(reduce_prod.symbols)
                items = []
                if items_needed != 0:
                    stack, items = stack[:-items_needed], stack[-items_needed:]
                items = [item[0].ast for item in items]

                # Apply reduction
                new_head = reduce_prod.head.copy()
                new_head.set_ast(reduce_prod.build_ast(items))
                logging.info(f"Reduced to {new_head}")

                # Check next state
                left_state = stack[-1][1] if stack else 0
                next_state = table[left_state, reduce_prod.head.name]
                logging.info(
                    f"Next state GOTO({left_state},{reduce_prod.head.name})"
                    f" is {next_state}"
                )
                # Push to stack the new item
                stack.append((new_head, next_state))
            elif table_val is None:
                raise ParsingError(f"Unexpected token", token)

        if len(stack) != 1:
            raise ValueError(f"Dirty stack at the end of the parsing. Stack: {stack}")
        return stack[-1][0].ast



"""
This module contains the exceptions used on the project.
"""


class NumlabError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class TokenizationError(NumlabError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class ParsingError(NumlabError):
    """
    This class represents an error that occurs during the parsing of a
    program.
    """

    def __init__(self, message, token):
        super().__init__(message)
        self.message = message
        self.token = token
        self.line = token.line
        self.column = token.col

    def __str__(self):
        return "{} at line {} column {}\n{}".format(
            self.message, self.line, self.column, self.token
        )


class RuntimeError(NumlabError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class ValueError(RuntimeError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class InvalidTypeError(RuntimeError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class InvalidConfigError(RuntimeError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class SimulationError(RuntimeError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message



import math
from typing import Callable


class FuzzyRule:
    def __init__(self, name: str, functions: list, rule: Callable):
        self.name = name
        self.functions = functions
        self.rule = rule

    def apply(self, *input_values: tuple):
        if len(input_values) != len(self.functions):
            raise ValueError(
                "The number of input values does not match the number of functions"
            )
        results = [self.functions[i](val) for i, val in enumerate(input_values)]
        return self.rule(*results)


class FuzzyOptClassifier:
    def __init__(self):
        self.max_loop_depth = 0
        self.changes = []
        self.rules = [
            FuzzyRule(
                name="Poor change",
                functions=[
                    lambda x: self._norm_bell_curve(0, 0.2, x),
                    lambda x: self._norm_bell_curve(0, 1, x),
                ],
                rule=max,
            ),
            FuzzyRule(
                name="Acceptable change",
                functions=[
                    lambda x: self._norm_bell_curve(0.5, 0.2, x),
                    lambda x: self._norm_bell_curve(0.5, 1, x),
                ],
                rule=max,
            ),
            FuzzyRule(
                name="Good change",
                functions=[
                    lambda x: self._norm_bell_curve(1, 0.2, x),
                    lambda x: self._norm_bell_curve(1, 1, x),
                ],
                rule=max,
            ),
        ]
        self.membership_val_centers = [0.1, 0.5, 1]

    def add_change(self, category: float, loop_depth: float):
        self.max_loop_depth = max(self.max_loop_depth, loop_depth)
        self.changes.append((category, loop_depth))

    def _loop_depth_val(self, loop_depth):
        return min(3, loop_depth)

    def _defuzzify(self, *values):
        return sum(
            [self.membership_val_centers[i] * values[i] for i in range(len(values))]
        ) / sum(values)

    def _classify_single_change(self, category: float, loop_depth: float):
        vals = [
            self.rules[i].apply(category, self._loop_depth_val(loop_depth))
            for i in range(len(self.rules))
        ]
        return self._defuzzify(*vals)

    def classify_changes(self):
        if len(self.changes) == 0:
            raise ValueError("No changes have been added")
        return [self._classify_single_change(*vals) for vals in self.changes]

    def classify_optimization_quality(self):
        if len(self.changes) == 0:
            raise ValueError("No changes have been added")
        vals = self.classify_changes()
        return max(vals)

    def _bell_curve(self, mean: float, std: float, x: float):
        return (1 / (std * math.sqrt(2 * math.pi))) * math.exp(
            -((x - mean) ** 2) / (2 * std ** 2)
        )

    def _norm_bell_curve(self, mean: float, std: float, x: float):
        return self._bell_curve(mean, std, x) / self._bell_curve(mean, std, mean)



from clismo.compiler.generic_ast import AST
from clismo.compiler.grammar import (Grammar, NonTerminal, Production, Symbol,
                                     Terminal)
from clismo.compiler.parsers.parser import Parser
from clismo.compiler.parsers.lritem import LRItem
from clismo.compiler.parsers.lr1_parser import LR1Parser, LR1Table

from clismo.compiler.parser_manager import ParserManager

from clismo.compiler.terminal_set import TerminalSet
from clismo.compiler.tokenizer import Token, Tokenizer



"""
This module contains the AST structure for the compiler.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any, List, Optional

DUMP_INDENT = 3


class AST(metaclass=ABCMeta):
    __slots__ = ()

    def dump(self, indent: int = 0, last_field: str = None) -> None:
        """
        Prints the AST in a human-readable format.
        """
        last_field = f"{last_field}: " if last_field else ""
        class_name = self.__class__.__name__
        class_name = f"({class_name})" if last_field else f"{class_name}:"
        if hasattr(self, "show"):
            # pylint: disable=no-member
            print(f"{' ' * indent}{last_field}{self.show()}")
            return
        print(f"{' ' * indent}{last_field}{class_name}")
        for field in self.__slots__:
            value = getattr(self, field)
            if isinstance(value, list) and value:
                print(f"{' ' * (indent + DUMP_INDENT)}{field}: [")
                for item in value:
                    if isinstance(item, AST):
                        item.dump(indent + DUMP_INDENT * 2)
                    else:
                        print(f"{' ' * (indent + DUMP_INDENT)}{item}")
                print(f"{' ' * (indent + DUMP_INDENT)}]")
            elif isinstance(value, AST):
                value.dump(indent + DUMP_INDENT, field)
            elif value:
                print(f"{' ' * (indent + DUMP_INDENT)}{field}: {value}")



from __future__ import annotations

import enum
from typing import Any, List

from clismo.compiler.generic_ast import AST


class Operator(enum.Enum):
    ADD = enum.auto()
    SUB = enum.auto()
    MUL = enum.auto()
    DIV = enum.auto()
    MOD = enum.auto()
    POW = enum.auto()
    AND = enum.auto()
    OR = enum.auto()
    LSHIFT = enum.auto()
    RSHIFT = enum.auto()
    BIT_AND = enum.auto()
    BIT_XOR = enum.auto()
    BIT_OR = enum.auto()
    FLOORDIV = enum.auto()
    EQ = enum.auto()
    NOT_EQ = enum.auto()
    LT = enum.auto()
    LTE = enum.auto()
    GT = enum.auto()
    GTE = enum.auto()
    NOT = enum.auto()
    UADD = enum.auto()
    USUB = enum.auto()
    INVERT = enum.auto()


class Program(AST):
    __slots__ = ("stmts",)

    def __init__(self, stmts: List[ObjDef]):
        self.stmts = stmts


class ObjDef(AST):
    def __init__(
        self,
        obj_type: str,
        name: str,
        body: List[AST],
    ):
        self.obj_type = obj_type
        self.name = name
        self.body = body


class ClientDef(ObjDef):
    __slots__ = ("name", "body")

    def __init__(self, name, body):
        super().__init__("client", name, body)


class ServerDef(ObjDef):
    __slots__ = ("name", "body")

    def __init__(self, name, body):
        super().__init__("server", name, body)


class StepDef(ObjDef):
    __slots__ = ("name", "body")

    def __init__(self, name, body):
        super().__init__("step", name, body)


class SimulationDef(ObjDef):
    __slots__ = ("name", "body")

    def __init__(self, name, body):
        super().__init__("simulation", name, body)


class Attr(AST):
    __slots__ = ("name", "value")

    def __init__(self, name: str, value: Expr):
        self.name = name
        self.value = value


class Function(AST):
    __slots__ = ("name", "info", "body")

    def __init__(self, name: str, info: List[Name], body: List[Stmt]):
        self.name = name
        self.info = info
        self.body = body


class Stmt(AST):
    pass


class Assign(Stmt):
    __slots__ = ("name", "value", "decl")

    def __init__(self, name: str, value: Expr, decl: bool = False):
        self.name = name
        self.value = value
        self.decl = decl


class If(Stmt):
    __slots__ = ("cond", "then", "els")

    def __init__(self, cond: Expr, then: List[Stmt], els: List[Stmt] = None):
        self.cond = cond
        self.then = then
        self.els = els or []


class Return(Stmt):
    __slots__ = ("value",)

    def __init__(self, value: Expr):
        self.value = value


class Loop(Stmt):
    __slots__ = ("target", "start", "end", "step", "body")

    def __init__(
        self,
        target: str,
        body: List[Stmt],
        start: Expr = None,
        end: Expr = None,
        step: Expr = None,
    ):
        self.target = target
        self.body = body
        self.start = start
        self.end = end
        self.step = step


class EndLoop(Stmt):
    pass


class NextLoop(Stmt):
    pass


class Expr(AST):
    pass


class Call(Expr):
    __slots__ = ("name", "args")

    def __init__(self, name: str, args: List[Expr] = None):
        self.name = name
        self.args = args or []


class BinOp(Expr):
    __slots__ = ("left", "op", "right")

    def __init__(self, left: Expr, op, right: Expr):
        self.left = left
        self.op = op
        self.right = right


class UnaryOp(Expr):
    __slots__ = ("op", "expr")

    def __init__(self, op, expr: Expr):
        self.op = op
        self.expr = expr


class ListExpr(Expr):
    __slots__ = ("elements",)

    def __init__(self, elements: List[Expr]):
        self.elements = elements


class Name(Expr):
    __slots__ = ("name",)

    def __init__(self, name: str):
        self.name = name

    def show(self):
        return f"Name: {self.name}"


class Constant(Expr):
    __slots__ = ("value",)

    def __init__(self, value: Any):
        self.value = value

    def show(self):
        return f"Constant: {self.value}"



import abc
from itertools import accumulate
from random import choices, random
from time import time
from typing import Any, Tuple


class Metaheuristic(metaclass=abc.ABCMeta):
    def __init__(self, max_iter: int = 1000, minimize=True) -> None:
        self.max_iter = max_iter
        self.minimize = minimize
        self.best_solution: Any = None

    @abc.abstractmethod
    def eval(self, solution) -> float:
        pass

    def get_best(self, sol1, sol2) -> Any:
        ev1, ev2 = self.eval(sol1), self.eval(sol2)
        if (self.minimize and ev1 <= ev2) or (not self.minimize and ev1 >= ev2):
            return sol1
        return sol2

    @abc.abstractmethod
    def iteration(self) -> None:
        pass

    def brake_condition(self) -> bool:
        return False

    def run(self):
        t_0 = time()
        i = 0
        while not self.brake_condition() and i < self.max_iter:
            print(f"Iteration {i + 1} of {self.max_iter}")
            self.iteration()
            i += 1
        t_1 = time()
        exec_time = t_1 - t_0
        return self.best_solution, self.eval(self.best_solution), exec_time


class GeneticAlg(Metaheuristic, metaclass=abc.ABCMeta):
    def __init__(
        self,
        max_iter: int = 20,
        minimize=True,
        population_size=100,
        mutation_prob=0.01,
        best_selection_count=10,
        generate_new_randoms=5,
    ):
        super().__init__(max_iter, minimize)
        self.new_best_found = True
        self.mutation_prob = mutation_prob
        self.population_size = population_size
        self.new_randoms = generate_new_randoms
        print(f"Generating population of {self.population_size}")
        self.population = [self.get_random_solution() for _ in range(population_size)]
        print(f"Evaluating population")
        self.population = [(p, self.eval(p)) for p in self.population]
        self.population.sort(key=lambda p: p[1], reverse=not self.minimize)
        self.best_selection_count = best_selection_count

    @abc.abstractmethod
    def get_random_solution(self):
        pass

    @abc.abstractmethod
    def crossover(self, sol1, sol2) -> Tuple[Any, Any]:
        pass

    @abc.abstractmethod
    def mutate(self, sol) -> Any:
        pass

    def iteration(self) -> None:
        # First selection (bests)
        new_candidates = self.population[: self.best_selection_count]

        new_rnds = [self.get_random_solution() for _ in range(self.new_randoms)]
        new_rnds = [(p, self.eval(p)) for p in new_rnds]
        new_candidates.extend(new_rnds)

        # Crossover
        total = sum([p[1] for p in self.population])
        cum_weigths = list(accumulate([p[1] / total for p in self.population]))

        while len(new_candidates) < self.population_size:
            c1, c2 = choices(self.population, cum_weights=cum_weigths, k=2)
            c1, c2 = c1[0], c2[0]
            new_c1, new_c2 = self.crossover(c1, c2)
            new_candidates.append((new_c1, self.eval(new_c1)))
            if len(new_candidates) < self.population_size:
                new_candidates.append((new_c2, self.eval(new_c2)))

        # Mutation
        crossed_start = self.best_selection_count + self.new_randoms
        for i in range(crossed_start, len(new_candidates)):
            rnd = random()
            if rnd < self.mutation_prob:
                candidate = new_candidates[i][0]
                mutaded = self.mutate(candidate)
                mutaded_val = self.eval(mutaded)
                new_candidates[i] = (mutaded, mutaded_val)

        self.population = new_candidates
        self.population.sort(key=lambda p: p[1], reverse=not self.minimize)
        self.best_solution = self.population[0][0]
        print(f"Best solution: {self.population[0]}")
        times = [p[1] for p in self.population]
        avg = sum(times) / len(times)





from clismo.lang.type import (
    cs_client,
    cs_server,
    cs_simulation,
    cs_step,
    cs_bool,
    cs_float,
    cs_int,
    cs_none,
    cs_object,
    cs_str,
    cs_list_of_float,
    cs_list_of_int,
    cs_list_of_str,
    cs_list_of_bool,
    cs_list_of_client,
    cs_list_of_server,
    cs_list_of_step,
    cs_list_of_simulation,
)

from clismo.builtin.cs_built_in_functions import resolve



import operator

import clismo.builtin as builtin
import clismo.cs_ast as ast
from clismo.lang.type import Type
from clismo.lang.visitor import Visitor
from clismo.sim.client import Client
from clismo.sim.server import Server
from clismo.sim.simulation import Simulation
from clismo.sim.step import Step

# pylint: disable=function-redefined
# pylint: disable=missing-function-docstring
# pylint: disable=unpacking-non-sequence


OPERATOR_FUNC = {
    ast.Operator.ADD: operator.add,
    ast.Operator.SUB: operator.sub,
    ast.Operator.MUL: operator.mul,
    ast.Operator.DIV: operator.truediv,
    ast.Operator.POW: operator.pow,
    ast.Operator.MOD: operator.mod,
    ast.Operator.POW: operator.pow,
    ast.Operator.LSHIFT: operator.lshift,
    ast.Operator.RSHIFT: operator.rshift,
    ast.Operator.BIT_XOR: operator.xor,
    ast.Operator.BIT_AND: operator.and_,
    ast.Operator.BIT_OR: operator.or_,
    ast.Operator.FLOORDIV: operator.floordiv,
    ast.Operator.EQ: operator.eq,
    ast.Operator.NOT_EQ: operator.ne,
    ast.Operator.LT: operator.lt,
    ast.Operator.GT: operator.gt,
    ast.Operator.LTE: operator.le,
    ast.Operator.GTE: operator.ge,
}


def _get_type_example(type_):
    if type_ == builtin.cs_bool:
        return True
    if type_ == builtin.cs_int:
        return 1
    if type_ == builtin.cs_float:
        return 1.0
    if type_ == builtin.cs_str:
        return "str"
    if type_.type_name.startswith("cs_list"):
        return [1, 2, 3]
    if type_ == builtin.cs_client:
        return Client("test")
    if type_ == builtin.cs_server:
        return Server("test", lambda x: 1)
    if type_ == builtin.cs_simulation:
        return Simulation("test", [], time_limit=1)
    if type_ == builtin.cs_step:
        return Step("test")
    raise Exception(f"Unknown type of node {type_}")


def _get_built_in_type(val):
    if isinstance(val, bool):
        return builtin.cs_bool
    if isinstance(val, int):
        return builtin.cs_int
    if isinstance(val, float):
        return builtin.cs_float
    if isinstance(val, str):
        return builtin.cs_str
    raise Exception(f"Unknown type of constant {val}")


class SemanticChecker:
    visitor = Visitor().visitor

    def __init__(self, types):
        self.types = types
        self.context = {}
        self.attrs = {}
        self.global_vars = {
            "time": builtin.cs_float,
            "clients": builtin.cs_int,
            "self": builtin.cs_object,
            "current_client": builtin.cs_client,
        }
        self.return_type = None
        self.current_func_name = None
        self.objects = {
            "Clients": {"DefaultClient": Client("DefaultClient")},
            "Servers": {},
            "Steps": {},
            "Simulations": {},
        }

    def resolve(self, name):
        if name in self.context:
            return self.context[name]
        if name in self.attrs:
            return self.attrs[name]
        if name in self.global_vars:
            return self.global_vars[name]
        if name in self.types["Clients"]:
            return builtin.cs_client
        if name in self.types["Servers"]:
            return builtin.cs_server
        if name in self.types["Simulations"]:
            return builtin.cs_simulation
        if name in self.types["Steps"]:
            return builtin.cs_step
        raise Exception(f"Unknown variable: {name}")

    def define(self, name, type_):
        if name in self.global_vars:
            raise Exception(f"Can not redefine global variable: {name}")
        self.context[name] = type_

    def _get_list_type(self, node):
        first_type = self.visit(node.elements[0])
        for elem in node.elements[1:]:
            if self.visit(elem) != first_type:
                raise Exception("All elements of list must be of the same type")
        list_type = f"{first_type.type_name}_list"
        return Type.get(list_type)

    def _get_type(self, node):
        if isinstance(node, ast.Constant):
            val = node.value
            return _get_built_in_type(val)
        if isinstance(node, ast.ListExpr):
            return self._get_list_type(node)
        if isinstance(node, ast.Name):
            if node.name not in self.context:
                raise Exception("Unknown variable: " + node.name)
            return self.context[node.value]
        raise Exception(f"Unknown type of node {node}")

    @visitor
    def visit(self, node: ast.Program):
        for stmt in node.stmts:
            self.visit(stmt)
        if not self.objects["Simulations"]:
            raise Exception("There must exist at least one simulation object")

    def _check_function_body(self, func_name, info, body, attrs, obj):
        if obj == "Clients":
            obj_type = builtin.cs_client
        elif obj == "Servers":
            obj_type = builtin.cs_server
        elif obj == "Simulations":
            obj_type = builtin.cs_simulation
        elif obj == "Steps":
            obj_type = builtin.cs_step
        else:
            raise Exception(f"Unknown object type: {obj}")

        self.global_vars["self"] = obj_type
        self.current_func_name = func_name
        for attr_name, attr_val in attrs.items():
            self.attrs[attr_name] = self._get_type(attr_val)
        if func_name == "on_server_out":
            self.return_type = builtin.cs_none
        elif func_name == "attend_client":
            self.return_type = builtin.cs_float
        elif func_name == "arrive":
            self.return_type = builtin.cs_float
        elif func_name == "minimize":
            self.return_type = builtin.cs_float
        elif func_name == "possible":
            self.return_type = self.resolve(info[0])
        else:
            assert False, "Unreachable"
        for stmt in body:
            self.visit(stmt)

    def _get_attrs_and_functions(self, dict_):
        functions = {k: v for k, v in dict_.items() if isinstance(v, dict)}
        attrs = {k: v for k, v in dict_.items() if k not in functions}
        if any(k in attrs for k in self.global_vars):
            raise Exception("Can not redefine global variable as an attribute")
        return attrs, functions

    def _validate_possible_func(self, info, attrs):
        if len(info) == 0:
            raise Exception("No attribute specified on 'possible' function")
        if len(info) > 1:
            raise Exception("Too many attributes specified on 'possible' function")
        if info[0] not in attrs:
            raise Exception(
                f"Attribute '{info[0]}' (specified on 'possible' "
                f"function) does not exist"
            )

    def _set_attrs_type(self, attrs, type_, name):
        for attr_name, attr_val in attrs.items():
            attr_type = self.visit(attr_val)
            self.types[type_][name][attr_name] = attr_type

    @visitor
    def visit(self, node: ast.ClientDef):
        attrs, functions = self._get_attrs_and_functions(
            self.types["Clients"][node.name]
        )
        self._set_attrs_type(attrs, "Clients", node.name)
        self.objects["Clients"][node.name] = Client(node.name)
        if "on_server_out" in functions:
            val = functions.pop("on_server_out")
            for info, body in val.items():
                for server_name in info:
                    if server_name not in self.types["Servers"]:
                        raise TypeError(
                            f"{node.name} on_server_out refers to "
                            f"unknown server {server_name}."
                        )
                self._check_function_body("on_server_out", info, body, attrs, "Clients")
        if "possible" in functions:
            val = functions.pop("possible")
            for info, body in val.items():
                self._validate_possible_func(info, attrs)
                self._check_function_body("possible", info, body, attrs, "Clients")
        if functions:
            raise TypeError(
                f"{node.name} has unknown functions: {list(functions.keys())}"
            )

    @visitor
    def visit(self, node: ast.ServerDef):
        attrs, functions = self._get_attrs_and_functions(
            self.types["Servers"][node.name]
        )
        self._set_attrs_type(attrs, "Servers", node.name)
        self.objects["Servers"][node.name] = Server(node.name, None)
        if "attend_client" in functions:
            val = functions.pop("attend_client")
            if len(list(val.items())) != 1:
                raise TypeError("There must exist exactly one 'attend_client' function")
            for info, body in val.items():
                if info:
                    raise TypeError(
                        "'attend_client' function must not have any specifications"
                    )
                self._check_function_body("attend_client", [], body, attrs, "Servers")
        if "possible" in functions:
            val = functions.pop("possible")
            for info, body in val.items():
                self._validate_possible_func(info, attrs)
                self._check_function_body("possible", info, body, attrs, "Servers")
        if functions:
            raise TypeError(
                f"{node.name} has unknown functions: {list(functions.keys())}"
            )

    @visitor
    def visit(self, node: ast.StepDef):
        attrs, functions = self._get_attrs_and_functions(self.types["Steps"][node.name])
        self._set_attrs_type(attrs, "Steps", node.name)
        self.objects["Steps"][node.name] = Step(node.name)
        if "servers" not in attrs:
            raise TypeError(f"{node.name} does not have servers.")
        servers = attrs["servers"]
        if not isinstance(servers, ast.ListExpr):
            raise TypeError(f"{node.name} servers attribute must be a list.")
        for server in servers.elements:
            if not isinstance(server, ast.Name):
                raise TypeError(
                    f"{node.name} servers attribute must be a list of names."
                )
            if server.name not in self.types["Servers"]:
                raise TypeError(
                    f"{node.name} servers attribute refers to "
                    f"unknown server {server.name}."
                )
        if "possible" in functions:
            val = functions.pop("possible")
            for info, body in val.items():
                self._validate_possible_func(info, attrs)
                self._check_function_body("possible", info, body, attrs, "Steps")
        if functions:
            raise TypeError(
                f"{node.name} has unknown functions: {list(functions.keys())}"
            )

    @visitor
    def visit(self, node: ast.SimulationDef):
        attrs, functions = self._get_attrs_and_functions(
            self.types["Simulations"][node.name]
        )
        self._set_attrs_type(attrs, "Simulations", node.name)
        self.objects["Simulations"][node.name] = Simulation(node.name, None)

        # Check steps
        if "steps" not in attrs:
            raise TypeError(f"{node.name} does not have steps.")
        steps = attrs["steps"]
        if not isinstance(steps, ast.ListExpr):
            raise TypeError(f"{node.name} steps attribute must be a list.")
        for step in steps.elements:
            if not isinstance(step, ast.Name):
                raise TypeError(f"{node.name} steps attribute must be a list of names.")
            if step.name not in self.types["Steps"]:
                raise TypeError(
                    f"{node.name} steps attribute refers to "
                    f"unknown step {step.name}."
                )

        # Check mode
        if "mode" in attrs:
            mode = attrs["mode"]
            if not isinstance(mode, ast.Constant) and not isinstance(mode.value, str):
                raise TypeError(f"{node.name} mode attribute must be a string.")
            mode_val = mode.value
            if mode_val not in ["run", "optimize"]:
                raise TypeError(
                    f"{node.name} mode attribute must be 'run' or 'optimize'."
                )
            if mode_val == "optimize" and not "minimize" in functions:
                raise TypeError(
                    f"{node.name} mode is 'optimize' but does not have "
                    f"a minimize function."
                )

        # Check terminate condition
        if "time_limit" not in attrs and "client_limit" not in attrs:
            raise TypeError(
                f"{node.name} does not have a time_limit or client_limit attribute."
            )
        if "time_limit" in attrs:
            val = attrs["time_limit"]
            if not isinstance(val, ast.Constant) and not isinstance(val.value, float):
                raise TypeError(f"{node.name} time_limit attribute must be float.")
        if "client_limit" in attrs:
            val = attrs["client_limit"]
            if not isinstance(val, ast.Constant) and not isinstance(val.value, int):
                raise TypeError(f"{node.name} client_limit attribute must be int.")

        # Check arrive function
        if "arrive" not in functions:
            raise TypeError(f"{node.name} does not have an arrive function.")
        arrive_overloads = functions.pop("arrive")
        for info, body in arrive_overloads.items():
            if not info:
                raise TypeError(
                    f"{node.name} arrive function must specify at least one "
                    "client type."
                )
            for client_type in info:
                if client_type not in self.types["Clients"]:
                    raise TypeError(
                        f"{node.name} arrive function refers to "
                        f"unknown client type {client_type}."
                    )
            self._check_function_body("arrive", info, body, attrs, "Simulations")

        # Check minimize function
        if "minimize" in functions:
            val = functions.pop("minimize")
            if len(list(val.keys())) > 1:
                raise TypeError(f"{node.name} minimize function has too many overloads")
            for info, body in val.items():
                if info:
                    raise TypeError(
                        f"{node.name} minimize function must have no specifications"
                    )
                self._check_function_body("minimize", info, body, attrs, "Simulations")

        if "possible" in functions:
            val = functions.pop("possible")
            for info, body in val.items():
                self._validate_possible_func(info, attrs)
                self._check_function_body("possible", info, body, attrs, "Simulations")
        if functions:
            raise TypeError(
                f"{node.name} has unknown functions: {list(functions.keys())}"
            )

    @visitor
    def visit(self, node: ast.Attr):
        pass

    @visitor
    def visit(self, node: ast.Function):
        pass

    @visitor
    def visit(self, node: ast.Assign):
        val_type = self.visit(node.value)
        if node.decl:
            self.define(node.name, val_type)
            return
        old_type = self.resolve(node.name)
        if not old_type.subtype(val_type):
            raise TypeError(
                f"{node.name} has type {old_type.type_name} but is being assigned "
                f"type {val_type.type_name}."
            )

    @visitor
    def visit(self, node: ast.If):
        cond_type = self.visit(node.cond)
        if not cond_type.subtype(builtin.cs_bool):
            raise TypeError(
                f"Condtion has type {cond_type.type_name} but must be bool."
            )
        for stmt in node.then:
            self.visit(stmt)

    @visitor
    def visit(self, node: ast.Return):
        val_type = self.visit(node.value)
        if not val_type.subtype(self.return_type):
            raise TypeError(
                f"Return value has type {val_type.type_name} but must "
                f"be {self.return_type.type_name}."
            )

    @visitor
    def visit(self, node: ast.Loop):
        if node.target is not None:
            self.define(node.target, builtin.cs_int)
        start, end, step = node.start, node.end, node.step
        if start is not None:
            start_type = self.visit(start)
            if not start_type.subtype(builtin.cs_int):
                raise TypeError(
                    f"Loop start has type {start_type.type_name} but must be int."
                )
        if end is not None:
            end_type = self.visit(end)
            if not end_type.subtype(builtin.cs_int):
                raise TypeError(
                    f"Loop end has type {end_type.type_name} but must be int."
                )
        if step is not None:
            step_type = self.visit(step)
            if not step_type.subtype(builtin.cs_int):
                raise TypeError(
                    f"Loop step has type {step_type.type_name} but must be int."
                )
        for stmt in node.body:
            self.visit(stmt)

    @visitor
    def visit(self, node: ast.EndLoop):
        pass

    @visitor
    def visit(self, node: ast.NextLoop):
        pass

    @visitor
    def visit(self, node: ast.Call):
        args_type = [self.visit(arg) for arg in node.args]
        if node.name == "get":
            if len(node.args) != 2:
                raise TypeError(
                    f"get function must have exactly two arguments, "
                    f"but got {len(node.args)}"
                )
            if not isinstance(node.args[1], ast.Constant) and not isinstance(
                node.args[1].value, str
            ):
                raise TypeError(
                    "get function second argument must be a constant string"
                )
            obj_type = self.visit(node.args[0])
            if obj_type == builtin.cs_server:
                type_ = "Servers"
            elif obj_type == builtin.cs_simulation:
                type_ = "Simulations"
            elif obj_type == builtin.cs_client:
                type_ = "Clients"
            elif obj_type == builtin.cs_step:
                type_ = "Steps"
            else:
                raise TypeError(
                    f"get function first argument must be a server, "
                    f"simulation, client, or step, but got {obj_type}"
                )
            attr_name = node.args[1].value
            for val in self.types[type_].values():
                if attr_name in val:
                    return val[attr_name]
            raise TypeError(
                f"{type_} {attr_name} does not exist."
            )
        if node.name == "set":
            if len(node.args) != 3:
                raise TypeError(
                    f"set function must have exactly three arguments, "
                    f"but got {len(node.args)}"
                )
            if not isinstance(node.args[1], ast.Constant) and not isinstance(
                node.args[1].value, str
            ):
                raise TypeError(
                    "set function second argument must be a "
                    "constant string (the attribute name)"
                )
            return builtin.cs_none
        ret_type = builtin.resolve(node.name)
        if ret_type is None:
            raise TypeError(f"{node.name} is not a known function.")
        if isinstance(ret_type[1], Type):
            return ret_type[1]
        if node.name == "list":
            if len(node.args) != 1:
                raise TypeError(
                    f"list function must have exactly one argument, "
                    f"but got {len(node.args)}"
                )
            if not isinstance(node.args[0], ast.Constant) and not isinstance(
                node.args[0].value, str
            ):
                raise TypeError(f"list function argument must be a constant string")
            return ret_type[1](node.args[0].value)
        return ret_type[1](*args_type)

    @visitor
    def visit(self, node: ast.BinOp):
        left_type = self.visit(node.left)
        right_type = self.visit(node.right)
        oper = OPERATOR_FUNC[node.op]
        exmp_left = _get_type_example(left_type)
        exmp_right = _get_type_example(right_type)
        try:
            new_type = _get_built_in_type(oper(exmp_left, exmp_right))
        except Exception:
            raise TypeError(
                f"{node.op} cannot be applied to types {left_type.type_name} and "
                f"{right_type.type_name}."
            )
        return new_type

    @visitor
    def visit(self, node: ast.UnaryOp):
        return self.visit(node.expr)

    @visitor
    def visit(self, node: ast.ListExpr):
        return self._get_list_type(node)

    @visitor
    def visit(self, node: ast.Name):
        return self.resolve(node.name)

    @visitor
    def visit(self, node: ast.Constant):
        return _get_built_in_type(node.value)



class OptimizableObject:
    def __init__(self, *childrens_names):
        self.posible_changes = []
        self.childrens_names = list(childrens_names) or []

    def add_possible_change(self, func, attr_name):
        self.posible_changes.append((self, attr_name, func))

    def get_possible_changes(self):
        ans = self.posible_changes
        for child in self.childrens_names:
            child = getattr(self, child)
            if isinstance(child, OptimizableObject):
                ans += child.get_possible_changes()
            elif isinstance(child, list):
                for item in child:
                    if isinstance(item, OptimizableObject):
                        ans += item.get_possible_changes()
        return ans



"""
This module contains the necessary functions to build an automata
from a regular expression and match a text against it.

Exmaple:
    >>> from regex_atmt import match
    >>> match("a*b", "ab")
    RegexMatch(matched:'ab'; end=2)"
    >>> match("a+b", "b")
    >>> match("a*b", "abb")
    RegexMatch(matched:'abb'; end=3)"
    >>> match("a*b", "abbfoo")
    RegexMatch(matched:'abb'; end=3)"
"""

import logging

from clismo.automata import Automata

ASCII = list(map(chr, range(128)))
SPECIAL_CHARS = ["(", ")", "|", "*", "^"]
DIGITS = list("0123456789")
LOWER_CASE_CHARS = list("abcdefghijklmnopqrstuvwxyz")
UPPER_CASE_CHARS = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")


class RegexMatch:
    """
    Represents the result of a match.

    Parameters
    ----------
    re_expr : str
        The regular expression used to match the text.
    text : str
        The text to match against the regular expression.
    pos : int
        The position where the match ended.

    Attributes
    ----------
    re_expr : str
        The regular expression used to match the text.
    text : str
        The text that was matched against the regular expression.
    pos : int
        The position where the match ended.
    """

    def __init__(self, re_expr: str, text: str, end: int):
        self.re_expr = re_expr
        self.text = text
        self.end = end

    def __repr__(self):
        return f"RegexMatch(matched:{self.text[:self.end]}; end={self.end})"

    @property
    def matched_text(self) -> str:
        return self.text[: self.end]


class RegexPattern:
    """
    Regular expression compiled to an automata.

    Parameters
    ----------
    re_expr : str
        The regular expression to be converted.

    Attributes
    ----------
    re_expr : str
        The regular expression.
    atmt : Automata
        The automata representing the regular expression.

    Raises
    ------
    ValueError
        If the regular expression is not valid.
    """

    def __init__(self, re_expr: str):
        self.re_expr = re_expr
        self.atmt = _build_automata(self.re_expr).flat().to_dfa()

    def match(self, text: str) -> RegexMatch:
        """
        Match the text against the regular expression.

        Parameters
        ----------
        text : str
            The text to match against the regular expression.

        Returns
        -------
        RegexMatch
            The RegexMatch object containing the result of the match.
        """

        last_pos = -1

        def set_last_pos():
            nonlocal last_pos
            last_pos = self.atmt.pos

        for state in self.atmt.end_states:
            state.on_visited = set_last_pos
        self.atmt.run(text)
        if last_pos == -1:
            return None
        return RegexMatch(self.re_expr, text, last_pos)


def _find_matching_paren(text: str, start: int = 0) -> int:
    count = 0
    for i in range(start, len(text)):
        if text[i] == "(":
            count += 1
        elif text[i] == ")":
            count -= 1
        if count == 0:
            return i
    raise ValueError("Unmatched parenthesis.")


def _get_basic_re_expr(re_expr: str) -> str:
    """
    Converts a regular expression to a basic regular expression.

    Parameters
    ----------
    re_expr : str
        Regular expression.

    Returns
    -------
    str
        Basic regular expression.
    """
    return re_expr


def _apply_star_op(atmt: Automata) -> Automata:
    """
    Applies the star operator to an automata.

    Parameters
    ----------
    atmt : Automata
        Automata.

    Returns
    -------
    Automata
        Automata.
    """
    #
    #           .---- < ----.
    #          /             \
    #  - - > (q0) -- .. --> (qf) --> ((new_qf))
    #          \                      /
    #           `-------- > ---------'
    #

    flat_atmt = atmt.flat()
    q_0 = flat_atmt.start_state
    q_f = flat_atmt.end_state
    flat_atmt.states.pop(q_f.name)
    for state in list(flat_atmt.states.values()):
        for trans in state.transitions:
            if trans.to_state == q_f:
                trans.to_state = q_0
    flat_atmt.end_states = [flat_atmt.start_state]
    return flat_atmt


def _check_special_char(re_expr, index, atmt):
    if index < 0 or index >= len(re_expr) or re_expr[index] != "*":
        return atmt, False
    atmt = _apply_star_op(atmt)
    return atmt, True


def _process_single_char(re_expr: str, negated: bool) -> Automata:
    """
    Processes a single character.

    Parameters
    ----------
    re_expr : str
        Regular expression.

    Returns
    -------
    Automata
        Automata.
    """
    #
    # - - > (q0) -- a --> ((q1))
    #
    escaped = False
    if re_expr[0] == "\\":
        re_expr = re_expr[1:]
        logging.debug("Escaped character")
        escaped = True
    cond = re_expr[0]
    logging.debug(f"Parsing {cond}")

    if escaped:
        if cond == "s":
            cond = " "
        elif cond == "n":
            cond = "\n"
        elif cond == "t":
            cond = "\t"
        elif cond == "r":
            cond = "\r"
        elif cond == "d":
            cond = DIGITS
        elif cond == "a":
            cond = LOWER_CASE_CHARS
        elif cond == "A":
            cond = UPPER_CASE_CHARS
    elif cond == ".":
        cond = None

    if negated:
        cond = [c for c in ASCII if c not in cond]

    new_atmt = Automata()
    from_st = new_atmt.add_state("q0", start=True)
    to_st = new_atmt.add_state("q1", end=True)
    new_atmt.add_transition(from_st, to_st, cond, action=1)
    logging.debug(f"Condition: {new_atmt.q0.transitions[0].str_cond}")
    new_atmt, changed = _check_special_char(re_expr, 1, new_atmt)
    logging.debug(f"Especial char found after {changed}")
    new_index = 2 if changed else 1
    new_index += 1 if escaped else 0
    return new_atmt, new_index


def _process_group(re_expr: str, negated: bool) -> Automata:
    """
    Processes a group.

    Parameters
    ----------
    re_expr : str
        Regular expression.

    Returns
    -------
    Automata
        Automata.
    """
    logging.debug("Parsing group")
    close_paren_index = _find_matching_paren(re_expr)
    negated = re_expr[1] == "^"
    start_index = 2 if negated else 1
    new_atmt = _build_automata(re_expr[start_index:close_paren_index], negated=negated)

    new_atmt, changed = _check_special_char(re_expr, close_paren_index + 1, new_atmt)

    logging.debug(f"Especial char found after group {changed}")
    new_index = close_paren_index + 2 if changed else close_paren_index + 1
    return new_atmt, new_index


def _process_or_operator(a_atmt: Automata, b_atmt: Automata) -> Automata:
    """
    Processes an or operator.

    Parameters
    ----------
    re_expr : str
        Regular expression.

    Returns
    -------
    Automata
        Automata.
    """
    #
    #              .-- > -- (a0) --..--> (af) -- > --.
    #             /                                   \
    # - - > (new_q0)                                  ((new_qf))
    #             \                                   /
    #              `-- > -- (b0) --..--> (bf) -- > --'
    #
    new_atmt = Automata()
    new_atmt.add_state("q0", start=True)
    new_atmt.add_state("qf", end=True)
    new_atmt.add_transition(new_atmt.q0, a_atmt.start_state)
    new_atmt.add_transition(new_atmt.q0, b_atmt.start_state)
    new_atmt.add_transition(a_atmt.end_state, new_atmt.qf)
    new_atmt.add_transition(b_atmt.end_state, new_atmt.qf)
    return new_atmt


def _build_automata(
    re_expr: str, last_atmt: Automata = None, negated: bool = False
) -> Automata:
    """
    Builds an automata from a regular expression using the Thompson
    construction algorithm.

    Parameters
    ----------
    re_expr : str
        Regular expression.

    Returns
    -------
    Automata
        Automata.
    """

    if re_expr == "":
        if last_atmt:
            return last_atmt
        raise ValueError("Invalid regular expression.")

    logging.debug(f"Building automata for {re_expr}")

    # Parse a single character
    if re_expr[0] not in SPECIAL_CHARS or re_expr[0] == "\\":
        new_atmt, new_index = _process_single_char(re_expr, negated)
        if last_atmt:
            new_atmt = last_atmt.concatenate(new_atmt)
        return _build_automata(re_expr[new_index:], new_atmt, negated)

    # Parse a group
    if re_expr[0] == "(":
        new_atmt, new_index = _process_group(re_expr, negated)
        if last_atmt:
            new_atmt = last_atmt.concatenate(new_atmt)
        return _build_automata(re_expr[new_index:], new_atmt, negated)

    # Parse an or operator
    if re_expr[0] == "|":
        logging.debug("Parsing or operator")
        if not last_atmt:
            raise ValueError("Invalid regular expression.")
        a_atmt = last_atmt
        b_atmt = _build_automata(re_expr[1:], None, negated)
        new_atmt = _process_or_operator(a_atmt, b_atmt)
        return new_atmt

    raise ValueError("Invalid regular expression {re_expr}")


def compile_patt(re_expr: str) -> RegexPattern:
    """
    Compiles a regular expression into an automata.

    Parameters
    ----------
    re_expr : str
        Regular expression.

    Returns
    -------
    Automata
        Automata.
    """
    return RegexPattern(re_expr)


def check(re_expr: str, text: str) -> bool:
    """
    Checks a regular expression against a text.

    Parameters
    ----------
    re_expr : str
        Regular expression.
    text : str
        Text.

    Returns
    -------
    bool
        True if the regular expression matches the text, False otherwise.
    """
    re_expr = _get_basic_re_expr(re_expr)
    re_patt = _build_automata(re_expr).flat()
    return re_patt.run(text)


def match(re_expr: str, text: str):
    """
    Matches a regular expression against a text.

    Parameters
    ----------
    re_expr : str
        Regular expression.
    text : str
        Text.

    Returns
    -------
    Match
        Match.
    """
    re_expr = _get_basic_re_expr(re_expr)
    re_patt = _build_automata(re_expr).flat().to_dfa()
    last_pos = -1

    def set_last_pos():
        nonlocal last_pos
        last_pos = re_patt.pos

    for state in re_patt.end_states:
        state.on_visited = set_last_pos
    re_patt.run(text)
    if last_pos == -1:
        return None
    return RegexMatch(re_expr, text, last_pos)



from __future__ import annotations

from typing import Any, Callable, List

from clismo.automata.transition import Transition


class State:
    """
    A state of an automata.

    Attributes
    ----------
    name : str
        The name of the state.
    transitions : List[Transition]
        The transitions of the state.
    """

    def __init__(self, name: str, on_visited: Callable = None) -> None:
        self.name = name
        self.transitions: List[Transition] = []
        self.automata = None
        self.on_visited = on_visited

    def visited(self):
        """
        Calls the ``on_visited`` callback if it is defined.
        """
        if self.on_visited:
            self.on_visited()

    def substitute(self, other_state: State) -> State:
        """
        Substitute the state with another one.

        Parameters
        ----------
        other_state : State
            The other state.

        Returns
        -------
        State
            The substituted state.
        """

        self.name = other_state.name
        self.transitions = other_state.transitions
        return self

    def merge(self, other_state: State) -> State:
        """
        Merge the state with another one.

        Parameters
        ----------
        other_state : State
            The other state.

        Returns
        -------
        State
            The merged state.
        """

        for trans in other_state.transitions:
            to_state = self if trans.to_state is other_state else trans.to_state
            new_t = Transition(
                self,
                to_state,
                trans.condition,
                trans.action,
                trans.negated,
            )
            self.transitions.append(new_t)
        return self

    def next_state(self, cond: Any):
        """
        Get the next state given a condition.

        Parameters
        ----------
        cond : Any
            The condition.

        Returns
        -------
        State
            The next state.
        """

        for trans in self.transitions:
            if trans.check_condition(cond):
                return trans.to_state
        return None

    def show(self) -> None:
        """
        Show the state.
        """

        print(self)
        for trans in self.transitions:
            print(f"  {trans}")

    def copy(self) -> State:
        """
        Copy the state.

        Returns
        -------
        State
            The copied state.
        """

        new_state = State(self.name)
        new_state.transitions = self.transitions
        return new_state

    def __str__(self) -> str:
        return self.name

    def __repr__(self):
        return str(self)



from clismo.sim.optimizable_obj import OptimizableObject


class Client(OptimizableObject):
    def __init__(self, name, **attrs):
        super().__init__()
        self.name = name
        self.attrs = attrs
        self.on_server_out_funcs = {}

    def add_on_server_out_callback(self, func, server: str):
        func_list = self.on_server_out_funcs.get(server, [])
        func_list.append(func)

    def on_server_out(self, server, time):
        func_list = self.on_server_out_funcs.get(server.name, [])
        for func in func_list:
            func(self, server, time)

    def get(self):
        cli = Client(self.name, **self.attrs.copy())
        cli.on_server_out_funcs = self.on_server_out_funcs
        return cli

    def __lt__(self, other):
        return self.name < other.name

    def __repr__(self):
        return self.name



import random
from contextlib import redirect_stdout
from time import time

from clismo.ia.genetic_alg import GeneticAlg
from clismo.sim.simulation import Simulation


class ModelOptimizer(GeneticAlg):
    """
    Genetic algorithm for optimizing the code
    """

    def __init__(
        self,
        model: Simulation,
        max_iter: int = 5,
        minimize=True,
        population_size=10,
        mutation_prob=0.1,
        best_selection_count=3,
        generate_new_randoms=2,
    ):
        self.model = model
        self.changes = model.get_possible_changes()
        super().__init__(
            max_iter,
            minimize,
            population_size,
            mutation_prob,
            best_selection_count,
            generate_new_randoms,
        )

    def _apply_solution(self, sol):
        """
        Apply a solution to the model
        """
        for obj, attr_name, value in sol:
            if hasattr(obj, attr_name):
                setattr(obj, attr_name, value)
            else:
                obj.attrs[attr_name] = value

    def eval(self, solution) -> float:
        """
        Evaluate the fitness of the solution
        """
        self._apply_solution(solution)
        self.model.run()
        return self.model.minimize_func()

    def get_random_solution(self):
        """
        Generate a random solution
        """
        return [(obj, attr_name, func()) for obj, attr_name, func in self.changes]

    def crossover(self, sol1, sol2):
        """
        Crossover two solutions
        """
        idx = random.randint(0, len(sol1) - 1)
        return sol1[:idx] + sol2[idx:], sol2[:idx] + sol1[idx:]

    def mutate(self, sol):
        """
        Mutate a solution
        """
        idx = random.randint(0, len(sol) - 1)
        obj, attr_name, func = self.changes[idx]
        sol[idx] = (obj, attr_name, func())
        return sol



import abc
from typing import List

from clismo.compiler.generic_ast import AST
from clismo.compiler.grammar import Grammar
from clismo.compiler.tokenizer import Token


class Parser(metaclass=abc.ABCMeta):
    def __init__(self, grammar: Grammar):
        self.grammar = grammar

    @abc.abstractmethod
    def parse(self, tokens: List[Token]) -> AST:
        pass



import builtins
import math
import random
from time import sleep

import clismo.exceptions as excpt
from clismo.lang.type import Instance, Type

__BUILTINS = {}

cs_float = Type.get("float")
cs_int = Type.get("int")
cs_str = Type.get("str")
cs_bool = Type.get("bool")
cs_none = Type.get("none")
cs_list = Type.get("list")


def builtin_func(func_name, return_type):
    def deco(func):
        __BUILTINS[func_name] = (func, return_type)
        return func

    return deco


def resolve(func_name):
    return __BUILTINS.get(func_name, None)


# region Numeric Functions


@builtin_func("abs", cs_float)
def cs_abs(x: cs_float):
    if x.type.subtype(cs_float):
        return cs_float(builtins.abs(x.value))
    raise TypeError("abs() argument must be a float")


@builtin_func("bin", cs_str)
def cs_bin(x: cs_int):
    if x.type.subtype(cs_int):
        return cs_str(builtins.bin(x.value))
    raise TypeError("bin() argument must be an int")


@builtin_func("round", cs_float)
def cs_round(x, ndigits=None):
    if x.type.subtype(cs_float):
        if ndigits is None:
            return cs_float(builtins.round(x.value))
        if ndigits.type.subtype(cs_int):
            return cs_float(builtins.round(x.value, ndigits.value))
        raise TypeError("round() second argument must be an int")
    raise TypeError("round() argument must be a float")


@builtin_func("rand", cs_float)
def cs_rand():
    return cs_float(random.random())


@builtin_func("randint", cs_int)
def cs_randint(a, b):
    if a.type.subtype(cs_int) and b.type.subtype(cs_int):
        return cs_int(random.randint(a.value, b.value))
    raise TypeError("randint() arguments must be int")


@builtin_func("norm", cs_float)
def cs_norm():
    return cs_float(random.normalvariate(0, 1))


@builtin_func("sqrt", cs_float)
def cs_sqrt(x):
    if x.type.subtype(cs_float):
        return cs_float(math.sqrt(x.value))
    raise TypeError("sqrt() argument must be a float")


@builtin_func("log", cs_float)
def log(x, base):
    if x.type.subtype(cs_float):
        if base.type.subtype(cs_int):
            return cs_float(math.log(x.value, base.value))
        raise TypeError("log() second argument must be an int")
    raise TypeError("log() argument must be a float")


@builtin_func("log2", cs_float)
def log2(x):
    if x.type.subtype(cs_float):
        return cs_float(math.log2(x.value))
    raise TypeError("log2() argument must be a float")


@builtin_func("exp", cs_float)
def exp(x):
    if x.type.subtype(cs_float):
        return cs_float(math.exp(x.value))
    raise TypeError("exp() argument must be a float")


@builtin_func("ceil", cs_int)
def ceil(x):
    if x.type.subtype(cs_int):
        return cs_int(math.ceil(x.value))
    raise TypeError("ceil() argument must be an int")


@builtin_func("floor", cs_int)
def floor(x):
    if x.type.subtype(cs_int):
        return cs_int(math.floor(x.value))
    raise TypeError("floor() argument must be an int")


@builtin_func("sin", cs_float)
def sin(x):
    if x.type.subtype(cs_float):
        return cs_float(math.sin(x.value))
    raise TypeError("sin() argument must be a float")


@builtin_func("cos", cs_float)
def cos(x):
    if x.type.subtype(cs_float):
        return cs_float(math.cos(x.value))
    raise TypeError("cos() argument must be a float")


@builtin_func("tan", cs_float)
def tan(x):
    if x.type.subtype(cs_float):
        return cs_float(math.tan(x.value))
    raise TypeError("tan() argument must be a float")


# endregion


# region List Functions


@builtin_func("len", cs_int)
def cs_len(x):
    if x.type.subtype(cs_list):
        return cs_int(len(x.value))
    raise TypeError("len() argument must be a list")


@builtin_func("get_at", lambda x, i: Type.get(x.type_name[:-5]))
def cs_get_at(x, index):
    if x.type.subtype(cs_list):
        if index.type.subtype(cs_int):
            return x.value[index.value]
        raise TypeError("get_at() second argument must be an int")
    raise TypeError("get_at() argument must be a list")


@builtin_func("set_at", cs_none)
def cs_set_at(x, index, obj):
    if x.type.subtype(cs_list):
        if index.type.subtype(cs_int):
            x.value[index.value] = obj
            return cs_none()
        raise TypeError("set_at() second argument must be an int")
    raise TypeError("set_at() argument must be a list")


@builtin_func("append", lambda x, o: x)
def cs_append(x, obj):
    if x.type.subtype(cs_list):
        x.value.append(obj)
        return x
    raise TypeError("append() argument must be a list")


@builtin_func("list", lambda x: Type.get(f"{x}_list"))
def cs_new_list(type_):
    list_type = Type.get(f"{type_.value}_list")
    return Instance(list_type, [])


# endregion


# region String Functions


@builtin_func("startswith", cs_bool)
def cs_startswith(x, y):
    if x.type.subtype(cs_str) and y.type.subtype(cs_str):
        return x.value.startswith(y.value)
    raise TypeError("startswith() arguments must be strings")


@builtin_func("isdigit", cs_bool)
def cs_isdigit(x):
    if x.type.subtype(cs_str):
        return x.value.isdigit(x.value)
    raise TypeError("isdigit() argument must be a str")


@builtin_func("lower", cs_str)
def cs_lower(x):
    if x.type.subtype(cs_str):
        return x.value.lower(x.value)
    raise TypeError("lower() argument must be a str")


@builtin_func("capitalize", cs_str)
def cs_capitalize(x):
    if x.type.subtype(cs_str):
        return x.value.capitalize(x.value)
    raise TypeError("capitalize() argument must be a str")


# endregion



from clismo.compiler.grammar import NonTerminal, Production, Symbol


class LRItem:
    """
    This class represents a single item in the LR state machine.
    """

    def __init__(self, prod: Production, dot_pos: int, lookahead: str = None):
        """
        Initializes a new LRItem.

        Parameters
        ----------
        production : Production
            Production that will be used.
        dot_pos : int
            Position of the dot in the production.
        """
        self.prod = prod
        self.lah = lookahead
        self.dot_pos = dot_pos
        self._repr = None

    @property
    def at_symbol(self) -> Symbol:
        """
        Returns the symbol at the dot.

        Returns
        -------
        Symbol
            Symbol at the dot.
        """
        if self.dot_pos < len(self.prod.symbols):
            return self.prod.symbols[self.dot_pos]
        return None

    @property
    def at_non_terminal(self) -> NonTerminal:
        """
        Returns the non-terminal at the dot.

        Returns
        -------
        NonTerminal
            Non-terminal at the dot.
        """
        at_symbol = self.at_symbol
        if at_symbol is None or at_symbol.is_terminal:
            return None
        return at_symbol

    def __repr__(self):
        """
        Returns a string representation of the item.

        Returns
        -------
        str
            String representation of the item.
        """

        if self._repr is not None:
            return self._repr
        head = f"{self.prod.head} -> "
        body_before_dot = " ".join(str(i) for i in self.prod.symbols[: self.dot_pos])
        body_after_dot = " ".join(str(i) for i in self.prod.symbols[self.dot_pos :])
        body = f"{body_before_dot} . {body_after_dot}"
        if self.lah is not None:
            text = f"{head}{body} [{self.lah}]"
        else:
            text = f"{head}{body}"
        self._repr = f"LRItem({text})"
        return self._repr

    def __eq__(self, other):
        if isinstance(other, tuple):
            prod, dot, lah = other
        elif isinstance(other, LRItem):
            prod, dot, lah = other.prod, other.dot_pos, other.lah
        else:
            return False
        same_prod = self.prod == prod
        same_dot = self.dot_pos == dot
        same_lah = self.lah == lah
        return same_prod and same_dot and same_lah

    def __hash__(self):
        return hash(self.__repr__())

    def __lt__(self, other):
        return str(self.prod) < str(other.prod)



import logging

import clismo.csre

logging.basicConfig(
    level=logging.WARNING,
    format="%(asctime)s [%(levelname)s] %(message)s",
    datefmt="%H:%M:%S",
)

__all__ = ["csre"]

__version__ = "0.1.0"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['clismo',
 'clismo.automata',
 'clismo.builtin',
 'clismo.compiler',
 'clismo.compiler.parsers',
 'clismo.ia',
 'clismo.lang',
 'clismo.optimization',
 'clismo.sim',
 'clismo.visitors']

package_data = \
{'': ['*']}

install_requires = \
['typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['clismo = clismo.main:app']}

setup_kwargs = {
    'name': 'clismo',
    'version': '0.1.0',
    'description': 'Programing language designed por simulating and optimizing client-server like discret events models',
    'long_description': None,
    'author': 'Jorge Morgado Vega',
    'author_email': 'jorge.morgadov@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
This module contains the basic structures for parsing.
"""

from __future__ import annotations

from typing import List

from clismo.compiler.generic_ast import AST
from clismo.compiler.grammar import Grammar
from clismo.compiler.parsers.lr1_parser import LR1Parser
from clismo.compiler.parsers.parser import Parser
from clismo.compiler.tokenizer import Token, Tokenizer


class ParserManager:
    """Structure used for parsing a file, text or token list given a
    grammar and a tokenizer.

    Parameters
    ----------
    grammar : Grammar
        Grammar that will be use for parsing.
    tokenizer : Tokenizer
        Tokenizer that will be use for tokenize a given txt.
    parser : Parser
        Parser that will be use for parsing a given list of tokens.
    """

    def __init__(
        self, grammar: Grammar, tokenizer: Tokenizer = None, parser: Parser = None
    ):
        self.grammar = grammar
        self.tokenizer = tokenizer
        self.parser = LR1Parser(grammar) if parser is None else parser

    def parse_file(self, file_path: str) -> AST:
        """Opens a file and parses it contents.

        Parameters
        ----------
        file_path : str
            File path.

        Returns
        -------
        AST
            AST generated by the parser.
        """
        with open(file_path, "r", encoding="utf-8") as file:
            text = file.read()
        return self.parse(text)

    def _replace(self, text, old, new):
        """Replaces all occurrences of old with new in text."""
        return new.join(text.split(old))

    def parse(self, text: str) -> AST:
        """Parses a text.

        Parameters
        ----------
        text : str
            Text to be parsed.

        Returns
        -------
        AST
            AST generated by the parser.
        """
        text = self._replace(text, "\t", "    ")
        tokens = self.tokenizer.tokenize(text)
        return self.parse_tokens(tokens)

    def parse_tokens(self, tokens: List[Token]) -> AST:
        """Parses a list of tokens.

        Parameters
        ----------
        tokens : List[Token]
            List of tokens to be parsed.
        method : str
            Method used for parsing.

        Returns
        -------
        AST
            AST generated by the parser.
        """
        tokens += [Token("$", "$")]
        return self.parser.parse(tokens)



"""
This module contains the structs necessary to represent an automata.
"""
from __future__ import annotations

import logging
from typing import Any, Dict, Iterable, List, Set, Tuple, Union

from clismo.automata.state import State
from clismo.automata.transition import Transition

_ATMT_COUNT = 0


class Automata:
    """
    An automata.

    Parameters
    ----------
    name : str
        The name of the automata.

    Attributes
    ----------
    name : str
        The name of the automata.
    states : Dict[str, State]
        The states of the automata.
    start_states : List[State]
        The start states of the automata.
    end_states : List[State]
        The end states of the automata.
    """

    def __init__(self, name: str = None) -> None:
        if name is None:
            global _ATMT_COUNT
            name = f"atmt_{_ATMT_COUNT}"
            _ATMT_COUNT += 1
        self.name = name
        self.states: Dict[str, State] = {}
        self.start_states: List[State] = []
        self.end_states: List[State] = []

        self._pos = 0
        self._input = None
        self._current_state: State = None
        self._processes: List[Tuple[State, int]] = []
        self._processes_idx: int = 0

    def __getattr__(self, item: str) -> Any:
        if item in self.states:
            return self.states[item]
        raise AttributeError(f"No attribute {item}")

    @property
    def alphabet(self) -> Set[Tuple[Any, bool]]:
        """
        Get the alphabet of the automata.

        Returns
        -------
        List[Any]
            The alphabet of the automata.
        """

        alphabet = set()
        for state in self.states.values():
            for transition in state.transitions:
                if transition.is_epsilon:
                    continue
                if isinstance(transition.condition, str):
                    alphabet.add(transition.condition)
                else:
                    alphabet.update(transition.condition)
        return alphabet

    def concatenate(self, other: Automata, set_single: bool = False) -> Automata:
        """
        Concatenate the automata with another one.

        Parameters
        ----------
        other : Automata
            The other automata.
        set_single : bool, optional
            Whether to set the automata to have a single start and end state
            when needed, by default False.

        Returns
        -------
        Automata
            The concatenated automata.

        Raises
        ------
        ValueError
            If the current automata has multiple end states and ``set_single`` is
            False.
        ValueError
            If the other automata has multiple start states and ``set_single`` is
            False.
        """

        if len(self.end_states) != 1:
            if set_single:
                self.set_single_end()
            else:
                raise ValueError(f"Automata {self.name} has multiple end states.")
        if len(other.start_states) != 1:
            if set_single:
                other.set_single_start()
            else:
                raise ValueError(f"Automata {other.name} has multiple start states.")
        other = other.flat()
        other_first_state = other.start_state
        other_last_state = other.end_state
        self.end_state.merge(other_first_state)
        if other_last_state == other_first_state:
            other_last_state = self.end_state
        for state in other.states.values():
            for trans in state.transitions:
                if trans.to_state is other_first_state:
                    trans.to_state = self.end_state
        self.end_states = [other_last_state]
        return self

    @property
    def pos(self) -> int:
        """Position of the automata on the input"""
        return self._pos

    @property
    def start_state(self) -> State:
        """Get the start state of the automata."""
        if len(self.start_states) == 1:
            return self.start_states[0]
        raise ValueError("The automata has multiple start states.")

    @property
    def end_state(self) -> State:
        """Get the end state of the automata."""
        if len(self.end_states) == 1:
            return self.end_states[0]
        raise ValueError("The automata has multiple end states.")

    def add_state(
        self,
        state: Union[str, State] = None,
        start: bool = False,
        end: bool = False,
        name: str = None,
    ) -> State:
        """
        Add a state to the automata.

        Parameters
        ----------
        state : Union[str, State]
            The name of the state to add or the state itself.
        start : bool
            Whether the state is a start state.
        end : bool
            Whether the state is an end state.

        Returns
        -------
        State
            The added state.
        """

        if state is None:
            state = State(f"q{len(self.states)}")
        if isinstance(state, str):
            if state in self.states:
                raise ValueError(f"State {state} already exists.")
            state = State(state)
            state.automata = self
        name = name if name is not None else state.name
        self.states[name] = state
        if start:
            self.start_states.append(state)
        if end:
            self.end_states.append(state)
        return state

    def add_transition(
        self,
        from_state: Union[str, State],
        to_state: Union[str, State],
        condition: Any = None,
        action: int = None,
        negated: bool = False,
    ) -> None:
        """
        Add a transition to the automata.

        Parameters
        ----------
        from_state : Union[str, State]
            The state from which the transition starts.
        to_state : Union[str, State]
            The state to which the transition goes.
        condition : Any
            The condition under which the transition is taken.
        action : int
            The action to perform when the transition is taken.

        Raises
        ------
        ValueError
            If any of the states does not exist.
        """

        if isinstance(from_state, str):
            from_state = self.states.get(from_state, None)
            if from_state is None:
                raise ValueError(f"No state {from_state} defined.")
        if isinstance(to_state, str):
            to_state = self.states.get(to_state, None)
            if to_state is None:
                raise ValueError(f"No state {to_state} defined.")
        if action is None:
            action = 0 if condition is None else 1
        transition = Transition(from_state, to_state, condition, action, negated)
        from_state.transitions.append(transition)
        return transition

    def set_single_start(self) -> State:
        """
        Set the automata to have a single start state.

        Returns
        -------
        State
            The start state.
        """

        if len(self.start_states) == 1:
            return self.start_states[0]
        start_st = self.add_state(f"_start_{self.name}")
        for state in self.start_states:
            self.add_transition(start_st, state)
        self.start_states = [start_st]
        return start_st

    def set_single_end(self) -> State:
        """
        Set the automata to have a single end state.

        Returns
        -------
        State
            The end state.
        """

        if len(self.end_states) == 1:
            return self.end_states[0]
        end_st = self.add_state(f"_end_{self.name}")
        for state in self.end_states:
            self.add_transition(state, end_st)
        self.end_states = [end_st]
        return end_st

    def set_single_start_end(self) -> Tuple[State, State]:
        """
        Set the automata to have a single start and end state.

        Returns
        -------
        Tuple[State, State]
            The start and end state.
        """

        start_st = self.set_single_start()
        end_st = self.set_single_end()
        return start_st, end_st

    def flat(self) -> Automata:
        """
        Flatten the automata.

        Returns
        -------
        Automata
            The flattened automata.
        """

        flat = Automata(self.name)
        count = 0
        visited_states = []
        non_visited_states = self.start_states
        while non_visited_states:
            new_non_visited_states = []
            for state in non_visited_states:
                flat.add_state(
                    state,
                    state in self.start_states,
                    state in self.end_states,
                    name=f"q{count}",
                )
                state.name = f"q{count}"
                count += 1
                visited_states.append(state)
                for transition in state.transitions:
                    to_state = transition.to_state
                    if (
                        to_state not in visited_states
                        and to_state not in new_non_visited_states
                        and to_state not in non_visited_states
                    ):
                        new_non_visited_states.append(transition.to_state)

            non_visited_states = new_non_visited_states
        return flat

    def show(self) -> None:
        """
        Show the automata.
        """

        # Inverse name states dict
        inv_states = {v: k for k, v in self.states.items()}

        for name, state in self.states.items():
            print(name, f"Final: {state in self.end_states}")
            for transition in state.transitions:
                neg = "^" if transition.negated else ""
                print(
                    f"  ({neg}{transition.str_cond}) "
                    f"-> {inv_states[transition.to_state]}"
                )

    def _eps_closure_single(self, state: Union[str, State]) -> Set[State]:
        """
        Compute the epsilon closure of a single state.

        Parameters
        ----------
        state : Union[str, State]
            The state to compute the epsilon closure of.

        Returns
        -------
        Set[State]
            The epsilon closure of the state.

        Raises
        ------
        ValueError
            If the state does not exist.
        """

        if isinstance(state, str):
            if state not in self.states:
                raise ValueError(f"No state {state} defined.")
            state = self.states[state]
        visited = set()
        non_vsited = [state]
        while non_vsited:
            new_non_vsited = []
            for current_state in non_vsited:
                visited.add(current_state)
                for transition in current_state.transitions:
                    if transition.is_epsilon:
                        to_st = transition.to_state
                        if (
                            to_st not in visited
                            and to_st not in new_non_vsited
                            and to_st not in non_vsited
                        ):
                            new_non_vsited.append(to_st)
            non_vsited = new_non_vsited
        return visited

    def eps_closure(
        self, state: Union[str, State, Iterable[str], Iterable[State]]
    ) -> Set[State]:
        """
        Compute the epsilon closure of a state or a set of states.

        Parameters
        ----------
        state : Union[str, State, Iterable[str], Iterable[State]]
            The state or a list of states.

        Returns
        -------
        Set[State]
            The epsilon closure of the state or a set of states.

        Raises
        ------
        ValueError
            If any of the states does not exist.
        """

        if isinstance(state, (str, State)):
            return self._eps_closure_single(state)
        whole_closure = set()
        for current_state in state:
            whole_closure.update(self._eps_closure_single(current_state))
        return whole_closure

    def _goto_single(self, state: Union[str, State], symbol: str) -> Set[State]:
        """
        Compute the goto of a single state.

        Parameters
        ----------
        state : Union[str, State]
            The state to compute the goto of.
        symbol : str
            The symbol to compute the goto of.

        Returns
        -------
        Set[State]
            The goto of the state.

        Raises
        ------
        ValueError
            If the state does not exist.
        """

        if isinstance(state, str):
            if state not in self.states:
                raise ValueError(f"No state {state} defined.")
            state = self.states[state]
        answer = set()
        st_esp_closure = self.eps_closure(state)
        for current_state in st_esp_closure:
            for transition in current_state.transitions:
                if not transition.is_epsilon and transition.check_condition(symbol):
                    answer.add(transition.to_state)
        return answer

    def goto(
        self, state: Union[str, State, Iterable[str], Iterable[State]], symbol: str
    ) -> Set[State]:
        """
        Compute the goto of a state or a set of states.

        Parameters
        ----------
        state : Union[str, State, Iterable[str], Iterable[State]]
            The state or a list of states.
        symbol : str
            The symbol to compute the goto of.

        Returns
        -------
        Set[State]
            The goto of the state or a set of states.

        Raises
        ------
        ValueError
            If any of the states does not exist.
        """

        if isinstance(state, (str, State)):
            return self._goto_single(state, symbol)
        whole_goto = set()
        for current_state in state:
            whole_goto.update(self._goto_single(current_state, symbol))
        return whole_goto

    def to_dfa(self, dfa2nfa: bool = False) -> Union[Automata, Tuple[Automata, Dict]]:
        """
        Convert the automata to a DFA.

        Parameters
        ----------
        dfa2nfa : bool
            If True, the return value will be a tuple of the DFA and the dfa2nfa
            dictionary, otherwise only the DFA will be returned. By default, False.

        Returns
        -------
        Union[Automata, Tuple[Automata, Dict]]
            The DFA.
        """

        get_name = lambda states: "".join(sorted(x.name for x in states))
        alphabet = self.alphabet
        dfa = Automata(self.name)
        start_state = self.eps_closure(self.start_states)
        start_name = get_name(start_state)
        q_0 = dfa.add_state(start_name, start=True, end=start_state in self.end_states)
        dfa_to_nfa = {q_0: start_state}
        visited = set()
        non_visited = [q_0]
        while non_visited:
            new_non_visited = []
            for current_state in non_visited:
                if current_state in visited:
                    continue
                visited.add(current_state)
                for char in alphabet:
                    goto_states = self.goto(dfa_to_nfa[current_state], char)
                    if not goto_states:
                        continue
                    next_state = self.eps_closure(goto_states)
                    next_name = get_name(next_state)
                    if next_name not in dfa.states:
                        dfa_state = dfa.add_state(
                            next_name,
                            end=any(s in self.end_states for s in next_state),
                        )
                        dfa_to_nfa[dfa_state] = next_state
                        new_non_visited.append(dfa_state)
                    else:
                        dfa_state = dfa.states[next_name]
                    dfa.add_transition(current_state.name, next_name, char)
                    if next_state not in new_non_visited and next_state not in visited:
                        new_non_visited.append(dfa_state)
            non_visited = new_non_visited
        return dfa if not dfa2nfa else (dfa, dfa_to_nfa)

    def run(
        self,
        input_: Iterable,
        stop_at_end: bool = False,
    ) -> bool:
        """
        Run the automata on the given input.

        Parameters
        ----------
        input_ : Iterable
            The input to run the automata on.
        stop_at_end : bool
            Whether to stop the automata at the first end state encountered.
        success_at_full_input : bool
            Whether to consider the automata successful if the input is fully
            consumed.

        Returns
        -------
        bool
            Whether the automata succeeded.

        Raises
        ------
        ValueError
            If the automata has no start state.
        """

        if not self.start_states:
            raise ValueError("No start states defined.")
        self._pos = 0
        self._processes_idx = 0
        self._input = input_
        self._processes = [(st, self._pos) for st in self.start_states]
        while self._processes:
            stop = self._step()
            if self._current_state in self.end_states:
                if stop_at_end:
                    return True
            if stop:
                break
        else:
            return False
        logging.debug(f"Final {self._processes_idx} {self._processes}")
        return self._current_state in self.end_states

    def _step(self):
        self._current_state, self._pos = self._processes[self._processes_idx]
        self._current_state.visited()
        if self._pos > len(self._input):
            self._processes.pop(self._processes_idx)
            return False
        new_processes = 0
        logging.debug(f"{self._processes_idx} {self._processes}")

        for transition in self._current_state.transitions:
            if transition.is_epsilon or (
                0 <= self._pos < len(self._input)
                and transition.check_condition(self._input[self._pos])
            ):
                run_state = (transition.to_state, self._pos + transition.action)
                if new_processes == 0:
                    self._processes[self._processes_idx] = run_state
                else:
                    self._processes.append(run_state)
                new_processes += 1

        if not new_processes:
            self._processes.pop(self._processes_idx)

        if self._processes:
            self._processes_idx = (self._processes_idx + 1) % len(self._processes)

        if self._pos >= len(self._input) or self._pos < 0:
            return self._current_state in self.end_states
        return False



from typing import Dict, List

from clismo.compiler.grammar import Grammar, Terminal
from clismo.compiler.terminal_set import TerminalSet


def calculate_first(gm: Grammar) -> Dict[str, TerminalSet]:
    first = {expr.name: TerminalSet() for expr in gm.exprs}
    prod_first = {prod: TerminalSet() for _, prod in gm.all_productions()}

    change = True
    while change:
        change = False
        for expr, prod in gm.all_productions():
            for item in prod.symbols:
                if item.is_terminal:
                    change |= first[expr.name].add(item)
                    prod_first[prod].add(item)
                    break
                if item != expr:
                    change |= first[expr.name].update(first[item.name])
                    prod_first[prod].update(first[item.name])
                if "EPS" not in first[expr.name].terminals:
                    break
    return first


def calculate_follow(gm: Grammar, first: TerminalSet = None) -> Dict[str, TerminalSet]:
    """
    Calculate the follow sets for all non-terminals in the grammar.

    Parameters
    ----------
    gm : Grammar
        The grammar to calculate the follow sets for.
    first : TerminalSet, optional
        The first sets of the grammar. If not given, it will be calculated
        first.

    Returns
    -------
    Dict[str, TerminalSet]
        The follow sets for all non-terminals in the grammar.
    """

    # First is needed to calculate follow
    if first is None:
        first = calculate_first(gm)

    follow = {expr.name: TerminalSet() for expr in gm.exprs}
    follow[gm.start_expr.name].add(Terminal("$"))

    change = True
    while change:
        change = False
        for expr, prod in gm.all_productions():
            for i, item in enumerate(prod.symbols):
                next_item = prod.symbols[i + 1] if i + 1 < len(prod.symbols) else None
                if item.is_terminal:
                    continue
                if next_item is None:
                    change |= follow[item.name].update(follow[expr] - "EPS")
                elif next_item.is_terminal:
                    change |= follow[item.name].add(next_item)
                else:
                    change |= follow[item.name].update(first[next_item] - "EPS")
                    if "EPS" in first[next_item].terminals:
                        change |= follow[item.name].update(follow[expr] - "EPS")
    return follow



from queue import PriorityQueue

from clismo.sim.optimizable_obj import OptimizableObject
from clismo.sim.server import Server


class Simulation(OptimizableObject):
    def __init__(self, name, steps, time_limit=None, client_limit=None):
        super().__init__("steps", "client_types")
        self.name = name
        self.arrival_funcs = []
        self.client_types = []
        self.steps = steps
        self.time = 0
        self.clients = 0
        self.events = PriorityQueue()
        self.time_limit = time_limit
        self.client_limit = client_limit
        self.minimize_func = None
        self.attrs = {}

    def add_arrival_func(self, arrival_func, client):
        self.arrival_funcs.append((arrival_func, client))
        self.client_types.append(client)

    def __run_arrivals(self):
        for arrival_func, client_type in self.arrival_funcs:
            delta_time = arrival_func()
            client = client_type.get()
            self.events.put(
                (self.time + delta_time, delta_time, client, Server.ghost(), 0)
            )

    def run(self, verbose=False):
        if self.time_limit is None and self.client_limit is None:
            raise ValueError("Either time_limit or client_limit must be specified")
        if not self.arrival_funcs:
            raise ValueError("No arrival functions specified")
        self.time = 0
        self.clients = 0
        self.events = PriorityQueue()
        for step in self.steps:
            step.clients_queue = []
            for i, server in enumerate(step.servers):
                step.servers[i] = server.get()

        while True:
            if self.events.empty():
                self.__run_arrivals()
            time, delta_time, client, last_server, step = self.events.get()
            if step < len(self.steps) and verbose:
                print(
                    f"{round(time, 3):>10} {client.name} "
                    f"arrived at {self.steps[step].name}",
                )
            elif verbose:
                print(
                    f"{round(time, 3):>10} {client.name} out of system",
                )
            if self.time_limit is not None and time > self.time_limit:
                break
            self.time = time
            if step > 0:
                last_server.in_use = False
                client.on_server_out(last_server, delta_time)
                event_time, server = self.steps[step - 1].next_client()
                if event_time is not None:
                    self.events.put(
                        (self.time + event_time, event_time, client, server, step)
                    )
            if step == len(self.steps):
                self.clients += 1
                if self.client_limit is not None and self.clients >= self.client_limit:
                    break
                continue
            if step == 0:
                self.__run_arrivals()
            event_time, server = self.steps[step].receive_client(client)
            if event_time is not None:
                self.events.put(
                    (self.time + event_time, event_time, client, server, step + 1)
                )
        if verbose:
            print()

    def __repr__(self):
        return self.name
